;; based on coop-sf-model-constrained-no-stay-behind.lisp
(clear-all)

(require-extra "tracker")

;; Parameters and constants
(defparameter *hex-radius* 175) ; in pixels
(defparameter *fortress-radius* 40) ; in pixels
(defparameter *acceleration* 0.05) ; pixels/gametick^2
(defparameter *radian* (/ 180 pi))

(defparameter *sgd-alpha* 0.1) ; stochastic gradient descent learning rate
;; (defparameter *alpha* 0.2)
;; (defparameter *factor* 1)

(defparameter *points* 0)
(defparameter *game-points* '())
(defparameter *moment-by-moment-data* '())
;(defparameter *experiment-points* '())
;(defparameter *params* '())

(defparameter *shells* '())
(defparameter *missiles* '())
(defparameter *fortress-alive* t)
(defparameter *self-alive* t)

;; Functions
(defun vector-length (vector-coordinates)
  (sqrt (apply #'+ (mapcar (lambda (x) (* x x)) vector-coordinates))))

(defun sqr (x)
  (* x x))

(defun to-radians (angle)
  (/ angle *radian*))

(defun to-degs (angle)
  (* angle *radian*))

(defun angle (x1 y1 x2 y2)
  ;; angle of vector from first to second point?
  (let* ((run (- x2 x1))
         (rise (- y2 y1))
         (a (if (not (zerop run))  (* (atan (/ rise run)) *radian*))))
    (cond ((zerop run) (if (>= rise 0) 90 270))
          ((> run 0) (if (>= rise 0) a (+ 360 a)))
          (t (if (>= rise 0) (+ 180 a) (+ 180 a))))))

(defun angle-difference (alpha beta)
  (let ((diff (- alpha beta)))
    (cond ((and (<= 0 diff) (>= 360 diff)) diff)
          ((> 0 diff) (angle-difference (+ 360 alpha) beta))
          ((< 360 diff) (angle-difference (- alpha 360) beta)))))

(defun angle-difference2 (alpha beta)
  (let ((diff (angle-difference alpha beta)))
    (if (> diff 180)
	(- diff 360)
	diff)))

(defun angles-within (alpha beta interval)
  (or (<= (angle-difference alpha beta) interval) (<= (angle-difference beta alpha) interval)))

(defun angle-offset (ang1 ang2)
  ;;calculates the difference between 2 angles dealing with looping at 360
  (let* ((offset (mod (- ang1 ang2) 360)))
    (if (> offset 180) (- 360 offset) offset)))

(defun closest-angle (angle1 angle2)
  (let ((angle-diff (angle-difference angle1 angle2)))
    (if (> angle-diff 180) (- 360 angle-diff) angle-diff)))

(defun remove-every-other (lst)
  (if (< (length lst) 2)
      lst
      (cons (first lst) (remove-every-other (cddr lst)))))

(defun clockwisep (angle1 angle2)
  (>= (angle-difference angle1 angle2) 180))

(defun counterclockwisep (angle1 angle2)
  (< (angle-difference angle1 angle2) 180))

(defun acceleratingp (vx vy aim)
  (angles-within (angle 0 0 vx vy) aim 90))

(defun time-to-radius (radius x y vx vy time)
  (if (> time 500)
      time
      (let ((current-radius (vector-length (list x y))))
	(if (> current-radius radius)
	    time
	    (time-to-radius radius (+ x vx) (+ y vy) vx vy (+ time 1))))))

(defun default-chunks (lis) (eval (cons 'add-dm (mapcar #'(lambda (x) (list x 'isa 'chunk)) lis))))

(defun angle-toward-fortress (x y)
  (angle-difference (angle 0 0 x y) 180))

(defun avoid-fortress-angle (x y fortress-back-angle) ;; (x y vx vy fortress-back-angle)
  ;; target flight direction is towards back of fortress
  (let* ((ship-position-angle (angle 0 0 x y))
         (fortress-ship-relative (if (clockwisep fortress-back-angle ship-position-angle) 'clockwise 'counterclockwise))
         (target-flight-direction (if (equalp fortress-ship-relative 'clockwise) (angle-difference ship-position-angle 60) (angle-difference ship-position-angle -60)))
         )
    target-flight-direction))

(defun compute-aim (target-fd vx vy target-speed)
  ;; takes into account that the minimum change in speed
  ;; is ~ 0.3 pixels / tick (based on a tap thrust of 90 ms)
  (let* ((target-fd-rad (to-radians target-fd))
         (target-vx (* target-speed (cos target-fd-rad)))
         (target-vy (* target-speed (sin target-fd-rad)))
	 (flight-dir (angle 0 0 vx vy))
         (delta-vx (- target-vx vx))
         (delta-vy (- target-vy vy))
         (delta-speed (vector-length (list delta-vx delta-vy))))
    (if (angles-within target-fd flight-dir 10)
	flight-dir
	(if (> delta-speed 0.3)
            (angle-difference (to-degs (atan delta-vy delta-vx)) 360)
            (let* ((target-vx-new (* (+ target-speed 0.1) (cos target-fd-rad)))
		   (target-vy-new (* (+ target-speed 0.1) (sin target-fd-rad)))
		   (delta-vx-new (- target-vx-new vx))
		   (delta-vy-new (- target-vy-new vy)))
              (angle-difference (to-degs (atan delta-vy-new delta-vx-new)) 360))))))

(defun behind-angle (ship-x ship-y ship-vx ship-vy fortress-back-angle shooter-distance horizon)
  (let* ((pred-x (+ ship-x (* ship-vx horizon)))
	 (pred-y (+ ship-y (* ship-vy horizon)))
	 (distance-from-center (sqrt (+ (sqr pred-x) (sqr pred-y))))
	 (ship-angle (angle 0 0 pred-x pred-y))
         (current-minus-target (angle-difference ship-angle fortress-back-angle)))
    (cond
      ((< distance-from-center 60)
       (if (> current-minus-target 0)
           (angle-difference ship-angle -45)
           (angle-difference ship-angle 45)))
      ((<= current-minus-target 30)
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle 90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle 90))))))
      ((> current-minus-target 330)
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle -90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle -90))))))
      ((or (and (<= current-minus-target 90) (> current-minus-target 30) )
           (and (> current-minus-target 270) (<= current-minus-target 330)))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians fortress-back-angle)))
              (* shooter-distance (sin (to-radians fortress-back-angle)))))
      ((and (> current-minus-target 90) (<= current-minus-target 180))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle 90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle 90))))))
      ((and (> current-minus-target 180) (<= current-minus-target 270))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle -90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle -90)))))))))

(defun outside-angle (x y vx vy horizon target-speed)
  (let* ((x-pred (+ x (* horizon vx)))
         (y-pred (+ y (* horizon vy)))
         (radius (vector-length (list x-pred y-pred)))
         (vx-outward (* (/ x-pred radius) target-speed))
         (vy-outward (* (/ y-pred radius) target-speed))
         (flight-direction (angle 0 0 vx vy))
         (ship-angle (angle 0 0 x-pred y-pred))
	 )
    (if (< (time-to-radius *hex-radius* x-pred y-pred vx vy 0)
           (time-to-radius *hex-radius* x-pred y-pred vx-outward vy-outward 0))
        flight-direction
        ship-angle)))

(defun bait-inside-angle (x y vx vy bait-distance)
  ;; choose direction perpendicular to radius in direction of current velocity
  (let* ((current-fd (angle 0 0 vx vy))
	 (current-loc-angle (angle 0 0 x y))
	 (target-location-angle (if (angles-within current-fd (angle-difference current-loc-angle 90) 90)
				    (angle-difference current-loc-angle 90)
				    (angle-difference current-loc-angle -90)))
	 (target-x (* bait-distance (cos (to-radians target-location-angle))))
	 (target-y (* bait-distance (sin (to-radians target-location-angle)))))
    (angle x y target-x target-y)))

(defun bait-angle (x y vx vy other-x other-y shell-x shell-y bait-distance)
  (let* ((ship-angle (angle 0 0 x y)) 
         (ship-distance (vector-length (list x y))) 
         (target-angle-counterclockwise (angle-difference ship-angle -90)) 
         (target-angle-clockwise (angle-difference ship-angle 90)) 
         (shell-angle (angle 0 0 shell-x shell-y))
         (shell-relative-position (if (clockwisep shell-angle ship-angle) 'clockwise 'counterclockwise))
         (speed (vector-length (list vx vy))) 
         (movement-relative-direction (if (clockwisep (angle 0 0 (+ x vx) (+ y vy)) ship-angle) 'clockwise 'counterclockwise)) 
         (other-ship-angle (angle 0 0 other-x other-y)) 
         (other-relative-direction (if (clockwisep other-ship-angle ship-angle) 'clockwise 'counterclockwise)) ;counterclockwise, but just a matter of less than 1 degree
         (final-target-angle
           (if (not (equalp shell-x 1000))
               (if (equalp shell-relative-position 'clockwise)
                   target-angle-counterclockwise
                   target-angle-clockwise) ;; if missiles present, fly in direction away from missiles
               (if (> ship-distance *hex-radius*)
                   (if (< (angle-difference ship-angle other-ship-angle) 100)
		       (if (equalp other-relative-direction 'clockwise)
			   target-angle-counterclockwise
			   target-angle-clockwise)
		       (if (equalp movement-relative-direction 'clockwise)
			   target-angle-clockwise
			   target-angle-counterclockwise))
                   (if (> speed 0.1)
                       (if (equalp movement-relative-direction 'clockwise)
                           target-angle-clockwise
                           target-angle-counterclockwise) ;; if no missiles are moving, change in direction closer to flight direction
                       (if (equal (random 2) 1)
                           target-angle-clockwise
                           target-angle-counterclockwise)) ;; if not moving, pick randomly one of the two possible targets
                   ))))
    (angle x y
           (* bait-distance (cos (to-radians final-target-angle)))
           (* bait-distance (sin (to-radians final-target-angle))))))

(defun stay-inside-angle (x y vx vy bait-distance)
  ;; choose direction perpendicular to radius in direction of current velocity
  (let* ((current-fd (angle 0 0 vx vy))
         (current-loc-angle (angle 0 0 x y))
         (target-location-angle (if (angles-within current-fd (angle-difference current-loc-angle 90) 90)
                                    (angle-difference current-loc-angle 90)
                                    (angle-difference current-loc-angle -90)))
         (target-x (* bait-distance (cos (to-radians target-location-angle))))
         (target-y (* bait-distance (sin (to-radians target-location-angle)))))
    (angle x y target-x target-y)))

(defun values-equal (dimension current target)
  (case dimension
    (location (or (and (equalp target 'insideHex) (equalp current 'behindFortress))
		  (equalp current target)))
    (other-location (or (and (equalp target 'insideHex) (equalp current 'behindFortress))
			(and (equalp target 'insideHex) (equalp current 'atBoundary))
			(equalp current target)))
    (fortress-target (equalp current target))
    (fortress-back (equalp current target))))

(defun location (x y fortress-angle fortress-alive)
  (labels ((behind-fortress-p (fortress-angle ship-x ship-y)
	     (let ((distance-from-center (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and
		(angles-within fortress-angle (angle-toward-fortress ship-x ship-y) 35)
		(<= distance-from-center (- *hex-radius* 10))
		(>= distance-from-center *fortress-radius*))))

	   (at-fortress-p (ship-x ship-y fortress-alive)
	     (let ((radius (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and fortress-alive (< radius *fortress-radius*))))

	   (at-outer-p (ship-x ship-y)
	     (or (<= ship-x -349) (>= ship-x 349) (<= ship-y -309) (>= ship-y 309))))

    (let ((distance-from-center (sqrt (+ (sqr x) (sqr y)))))
      (cond
	((at-outer-p x y) 'crashedIntoOuter)
	((> distance-from-center (+ *hex-radius* 17)) 'outsideHex)
	((> distance-from-center (- *hex-radius* 10)) 'atBoundary)
	((at-fortress-p x y fortress-alive) 'crashedIntoFortress)
	((behind-fortress-p fortress-angle x y) 'behindFortress)
	((<= distance-from-center *hex-radius*) 'insideHex)))))

(defun predict-back-orientation (other-x other-y other-vx other-vy horizon)
  (let* ((anticip-other-x (+ other-x (* horizon other-vx)))
         (anticip-other-y (+ other-y (* horizon other-vy)))
         (other-angle (angle 0 0 anticip-other-x anticip-other-y)))
    (angle-difference other-angle 180)))

(defun predicted-location (x y vx vy ox oy ovx ovy fortress-alive horizon)
  (labels ((at-outer-p (ship-x ship-y)
	     (or (<= ship-x -349) (>= ship-x 349) (<= ship-y -309) (>= ship-y 309)))
	   (future-at-outer-p (ship-x ship-y ship-vx ship-vy)
	     (at-outer-p (+ ship-x (* ship-vx horizon))
			 (+ ship-y (* ship-vy horizon))))

	   (behind-fortress-p (fortress-angle ship-x ship-y)
	     (let ((distance-from-center (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and
		(angles-within fortress-angle (angle-toward-fortress ship-x ship-y) 35)
		(<= distance-from-center (- *hex-radius* 10))
		(>= distance-from-center *fortress-radius*))))

	   (future-behind-fortress-p (ship-x ship-y ship-vx ship-vy other-x other-y other-vx other-vy horizon)
	     (behind-fortress-p (angle-difference
				 (predict-back-orientation other-x other-y other-vx other-vy horizon) 180)
				(+ ship-x (* horizon ship-vx))
				(+ ship-y (* horizon ship-vy))))

	   (at-fortress-p (ship-x ship-y fortress-alive)
	     (let ((radius (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and fortress-alive (< radius *fortress-radius*))))

	   (future-at-fortress-p (ship-x ship-y ship-vx ship-vy fortress-alive)
	     (at-fortress-p (+ ship-x (* ship-vx horizon))
			    (+ ship-y (* ship-vy horizon))
			    fortress-alive)))
    (let* ((future-x (+ x (* horizon vx)))
           (future-y (+ y (* horizon vy)))
           (future-distance-from-center (vector-length (list future-x future-y))))
      (cond
	((future-at-outer-p x y vx vy) 'crashedIntoOuter)
	((> future-distance-from-center (+ *hex-radius* 17)) 'outsideHex)
	((> future-distance-from-center (- *hex-radius* 10)) 'atBoundary)
	((future-at-fortress-p x y vx vy fortress-alive) 'crashedIntoFortress)
	((and (equal fortress-alive 'yes) (future-behind-fortress-p x y vx vy ox oy ovx ovy horizon)) 'behindFortress)
	((<= future-distance-from-center *hex-radius*) 'insideHex)))))

(defun predicted-location-union (x y vx vy ox oy ovx ovy fortress-alive horizon)
  (let ((current-location (predicted-location x y vx vy ox oy ovx ovy fortress-alive 0))
	(next-location (predicted-location x y vx vy ox oy ovx ovy fortress-alive 1)))
    (if (or (not (equal current-location next-location))
	    (<= horizon 1))
	next-location
	(predicted-location-union (+ x vx) (+ y vy) vx vy (+ ox ovx) (+ oy ovy) ovx ovy fortress-alive (- horizon 1)))))

(defun predict-aim-at-fortress (x y vx vy horizon)
  (angle 0 0 (- (+ x (* horizon vx))) (- (+ y (* horizon vy)))))

(defun predict-direction (vx vy aim horizon)
  (angle 0 0 (+ vx (* *acceleration* horizon (cos (/ aim *radian*))))
	 (+ vy (* *acceleration* horizon (sin (/ aim *radian*))))))

(defun predict-speed(vx vy aim horizon)
  (vector-length (list (+ vx (* *acceleration* horizon (cos (/ aim *radian*))))
		      (+ vy (* *acceleration* horizon (sin (/ aim *radian*)))))))

(defun direction-close (current target predicted)
  (or
   ;(angles-within current target 10)
   (and (> current target) (< predicted target))
   (and (< current target) (> predicted target))))

(defun speed-close (current target)
  (< (abs (- current target)) 0.14))

(defun flight-direction-bait-p (x y vx vy)
  (let ((target (angle-difference (angle 0 0 x y) 90))
	(speed (vector-length (list vx vy)))
	(current (angle 0 0 vx vy)))
    (and
     (> speed 0.05)
     (or
      (angles-within target current 55)
      (angles-within (angle-difference target 180) current 55)))))

(defun aim-good (aim orientation aim-delta)
  (angles-within aim orientation aim-delta))

(defun drifting-away (x y vx vy orientation)
  (or (and (clockwisep (angle 0 0 (+ x vx) (+ y vy)) (angle 0 0 x y))
	   (counterclockwisep (angle 0 0 x y) orientation))
      (and (counterclockwisep (angle 0 0 (+ x vx) (+ y vy)) (angle 0 0 x y))
	   (clockwisep (angle 0 0 x y) orientation))))

;; A fuction to update state in imaginal for the purposes of providing feedback to tracker
(defun update-slot (slot)
  (schedule-mod-buffer-chunk 'imaginal (list slot 1) 0))

;; Populate game-state buffer
(defun game-state-features-coop (data)
  (if (equal (getf data :state) "score")
      (progn
        (if  (not (equal (chunk-slot-value-fct (buffer-read 'game-state) 'status) 'game-over))
             (push (getf data :POINTS) *game-points*))
        (mod-buffer-chunk 'game-state
                          (list
			   'game nil
			   'status 'game-over
                           'ship-alive nil
                           'x nil
                           'y nil
                           'vx nil
                           'vy nil
			   'speed nil
			   'flight-direction nil
			   'location nil
			   'orientation nil
			   'aim-at-fortress nil
			   'fortress-angle nil
                           'other-x nil
                           'other-y nil
			   'other-vx nil
			   'other-vy nil
			   'other-location nil
                           'shell-x nil
                           'shell-y nil
			   'missile-present nil
                           'fortress-alive nil
                           'fortress-target nil
                           'fortress-back nil )))
      (let* ((gamestate (getf data :GAMESTATE))
             (self-data (if (equal (get-parameter-value :sf-player) 0)
                            (first (getf gamestate :PLAYERS))
                            (second (getf gamestate :PLAYERS))))
             (other-data (if (equal (get-parameter-value :sf-player) 0)
                             (second (getf gamestate :PLAYERS))
                             (first (getf gamestate :PLAYERS))))
             (missiles (getf gamestate :Missiles))
             (shells (getf gamestate :SHELLS))
             (shell-x (getf (getf (first shells) :POS) :X))
             (shell-y (getf (getf (first shells) :POS) :Y))
             (fortress-data (getf gamestate :FORTRESS))
             (fortress-alive (if (getf fortress-data :alive) 'yes 'no))
             (x (getf (getf self-data :POS) :X))
             (y (getf (getf self-data :POS) :Y))
             (vx (getf (getf self-data :VEL) :X))
             (vy (getf (getf self-data :VEL) :Y))
             (orientation (getf self-data :ANGLE))
             (speed (vector-length (list vx vy)))
             (other-x (getf (getf other-data :POS) :X))
             (other-y (getf (getf other-data :POS) :Y))
             (other-vx (getf (getf other-data :VEL) :X))
             (other-vy (getf (getf other-data :VEL) :Y))
             (flight-direction (if (< speed 0.01) orientation (angle 0 0 vx vy))) ;; if stopped, FD is same as orientation
             (fortress-angle (getf fortress-data :ANGLE))
             (self-alive (if (getf self-data :ALIVE) 'yes 'no))
             ;(other-alive (if (getf other-data :ALIVE) 'yes 'no))
	     (fortress-target-self (if (equal (get-parameter-value :sf-player) '0)
                                       (if (equal (getf fortress-data :TARGET) 0) 'yes 'no)
                                       (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)))
	     (fortress-target-other (if (equal (get-parameter-value :sf-player) '0)
                                        (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)
                                        (if (equal (getf fortress-data :TARGET) 0) 'yes 'no))))
          (cond ((getf self-data ':ALIVE)
		 (mod-buffer-chunk 'game-state
                                   (list
				    'game 'coop
				    'status 'play
				    'ship-alive self-alive
				    'x x
				    'y y
				    'vx vx
				    'vy vy
				    'speed speed
				    'flight-direction flight-direction
				    'location (location x y fortress-angle fortress-alive)
				    'orientation orientation
				    'aim-at-fortress (angle-difference (angle 0 0 x y) 180)
				    'fortress-angle fortress-angle
				    'other-x other-x
				    'other-y other-y
				    'other-vx other-vx
				    'other-vy other-vy
				    'other-location (location other-x other-y fortress-angle fortress-alive)
				    'shell-x (if shells shell-x 1000)
				    'shell-y (if shells shell-y 1000)
				    'missile-present (if missiles 'yes 'no)
				    'fortress-alive fortress-alive
				    'fortress-target (cond ((equalp fortress-target-self 'yes) 'self)
							   ((equalp fortress-target-other 'yes) 'other)
							   (t 'none))
				    'fortress-back (if (getf fortress-data :VULNERABLE) 'vulnerable 'notvulnerable)
                                    )))
		(t
		 (mod-buffer-chunk 'game-state
				   (list
				    'game nil
				    'status nil
				    'ship-alive 'no
				    'x nil
				    'y nil
				    'vx nil
				    'vy nil
				    'speed nil
				    'flight-direction nil
				    'location nil
				    'orientation nil
				    'aim-at-fortress nil
				    'fortress-angle nil
				    'other-x nil
				    'other-y nil
				    'other-vx nil
				    'other-vy nil
				    'shell-x nil
				    'shell-y nil
				    'missile-present nil
				    'fortress-alive nil
				    'fortress-target nil
				    'fortress-back nil
				    )))))))

(defun update-theta (theta delta)
  ;; Function to learn parameters with stochastic gradient descent
  (- theta (* *sgd-alpha* delta)))

(defun record-features-coop (data)
  ;; (format t "Hook called with: ~s~%" data)
  (game-state-features-coop data)
  (if (not (equal (getf data :state) "score"))
      (let* ((gamestate (getf data :GAMESTATE))
             (points (getf gamestate ':POINTS))
             (self-data (if (equal (get-parameter-value :sf-player) 0)
			    (first (getf gamestate :PLAYERS))
			    (second (getf gamestate :PLAYERS))))
             (self-alive (if (getf self-data :ALIVE) 'yes 'no))
             ;; (x (getf (getf self-data :POS) :X))
             ;; (y (getf (getf self-data :POS) :Y))
             (fortress-data (getf gamestate :FORTRESS))
             (fortress-alive (if (getf fortress-data :alive) 'yes 'no))
             ;; (fortress-angle (getf fortress-data :ANGLE))
             (shells (getf gamestate :SHELLS))
	     ;; (missiles (getf gamestate :MISSILES))
	     (fortress-target-self (if (equal (get-parameter-value :sf-player) '0)
                                       (if (equal (getf fortress-data :TARGET) 0) 'yes 'no)
                                       (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)))
	     ) 
	(if (and (< (length *shells*) (length shells))
		 (equal fortress-target-self 'yes))
	    (update-slot 'shell-fired))
	;; (if (< (length *missiles*) (length missiles)) (progn
	;; 					       (update-slot 'missile-fired)
	;; 					       (format t "~a" "MISSILE FIRED"))
	;;     )
	(if (and (equal self-alive 'no) (equal *self-alive* 'yes)) (update-slot 'death))
	(if (and (equal fortress-alive 'no) (equal *fortress-alive* 'yes)) (update-slot 'kill))
	(if (and (< (length shells) (length *shells*)) (equalp fortress-alive 'yes)) (update-slot 'miss))
	(setf *self-alive* self-alive)
	(setf *fortress-alive* fortress-alive)
	(setf *shells* shells)
	;; (setf *missiles* missiles)
	(setf *points* points)
	)))

(defun get-server-port ()
  (let ((v (if (and (boundp '*sfcoop-headless-server-port*)
                    (symbol-value '*sfcoop-headless-server-port*))
               (symbol-value '*sfcoop-headless-server-port*)
               3000)))
    (format t "; server port is ~S~%" v)
    v))
  
(defun report-progress (c limit)
  (when (and (boundp '*cluster-progress-update-function*)
             (symbol-value '*cluster-progress-update-function*))
    (funcall (symbol-value '*cluster-progress-update-function*) (/ c limit))))

(defparameter *sgps*
  `((sgp
     :sf-data-hook record-features-coop
;;     :sf-port ,(get-server-port)
;;     :sf-game-count-hook report-progress 
     :v t
     :trace-detail low
     ;; declarative parameters
     :esc t
     :lf .1
     :ans .01 ; activation noise
     :rt -1 ; retrieval threshold
     :ol t ; optimized learning
     :er t ; randomly picks a chunk among multiple matching chunks when :esc is nil
     ;; utility learning parameters
     :ul t ;; utility learning switch
     :iu 9 ;; initial utility value of productions
     :egs .05 ;; utility noise parameter :egs .0001
     ;:alpha ,*alpha*
     ;; production compilation
     :epl t ; enables production compilation
     :pct nil ; determines whether information about production compilation is output in model trace
     :cia t
     :rir t
     ;:initial-temp ,*factor*

     ;; other parameters
     :randomize-time 3 ; randomizes perceptual and motor firing times
    ;; :test-feats nil
     :Visual-movement-tolerance 2
     :do-not-harvest visual
     ;; Temporal module's noise
     :TIME-NOISE 0.005

     ;; Motor setup, timing, and randomization
     :dual-execution-stages t
     :MOTOR-FEATURE-PREP-TIME 0.020
     :MOTOR-INITIATION-TIME .020
;     :hold-until-timing .025
     )))

(defparameter *action-to-key-map*
  '((thrust-key isa mapping function thrust key w)
    (counterclockwise-turn-key isa mapping function counterclockwise key a)
    (clockwise-turn-key isa mapping function clockwise key d)
    (fire-key isa mapping function shoot key spacebar)))

(defparameter *operators*
   '((crash-outer?
      isa operator
      pre coop
      action test-predicted-location
      arg1 crashedIntoOuter
      arg2 locationHorizon
      success thrust-stop
      fail wont-crash-outer)

     (aim-opposite-flight-direction
      isa operator
      pre thrust-opposite
      action aim-opposite
      post thrust-direction)

     (thrust-dir
      isa operator
      pre thrust-direction
      action thrust-to-turn
      post done)

     (fortress-alive?
      isa operator
      pre wont-crash-outer
      action test-object
      arg1 fortress-alive
      success fortress
      fail nofortress)

     (crash-fortress?
      isa operator
      pre fortress
      action test-predicted-location-union
      arg1 crashedIntoFortress
      arg2 locationHorizon
      success crash-fortress
      fail no-crash-fortress)

     (determine-avoid-fortress-angle
      isa operator
      pre crash-fortress
      action calculate-avoid-fortress-angle
      post check-flight-dir-slow)

     (determine-thrust-angle
      isa operator
      pre thrust-angle
      action calculate-thrust-angle
      arg1 good-speed
      post thrust-direction)

     (determine-thrust-angle-slow
      isa operator
      pre thrust-angle-slow
      action calculate-thrust-angle
      arg1 slow-speed
      post thrust-direction)

     (fortress-no-target?
      isa operator
      pre no-crash-fortress
      action test-categorical-dimension
      arg1 fortress-target
      arg2 none
      success notarget
      fail yestarget)

     (role?
      isa operator
      pre notarget
      action test-role
      success role
      fail norole)

     (update-role
      isa operator
      pre yestarget
      action updaterole
      post role)

     (select-role
      isa operator
      pre norole
      action pick-role
      post role)

     (role-bait?
      isa operator
      pre role
      action test-role
      arg1 bait
      success bait
      fail shooter)

     (self-inside?
      isa operator
      pre bait
      action test-categorical-dimension
      arg1 location
      arg2 insideHex
      success bait-inside
      fail bait-notinside)

     (self-direction-good?
      isa operator
      pre bait-inside
      action test-bait-direction
      success flight-direction-good
      fail flight-direction-bad)

     (determine-bait-direction
      isa operator
      pre flight-direction-bad
      action calculate-good-bait-direction
      post thrust-angle)

     (self-slow?
      isa operator
      pre flight-direction-good
      action test-speed
      arg1 slow-speed
      success bait-slow
      slow aim-speed-up
      fast aim-slow-down)

     (will-exit?
      isa operator
      pre bait-slow
      action test-predicted-location
      arg1 outsideHex
      arg2 avoidExitHorizon
      fail nothing-to-do
      success bait-will-exit)

     (future-inside?
      isa operator
      pre bait-notinside
      action test-predicted-location
      arg1 atBoundary
      arg2 aimHorizon
      success bait-future-inside
      fail bait-future-not-inside)

     (determine-bait-stay-inside-direction
      isa operator
      pre bait-future-not-inside
      action calculate-bait-direction
      post check-flight-dir)

     (anticipatory-turn
      isa operator
      pre bait-future-inside
      action aim-opposite
      post nothing-to-do)

     (set-speed-inside
      isa operator
      pre set-speed-slow-inside
      action set-speed
      arg1 slow-speed
      post thrust-slow-inside)

     (thrust-slow-when-inside
      isa operator
      pre thrust-slow-inside
      action thrust-speed
      arg1 slow-speed
      arg2 insideHex
      post done)

     (determine-stay-inside-direction
      isa operator
      pre bait-will-exit
      action calculate-stay-inside-direction
      post check-flight-dir-slow)

     (other-inside?
      isa operator
      pre shooter
      action test-categorical-dimension
      arg1 other-location
      arg2 insideHex
      success other-inside
      fail other-not-inside)

     (test-shooter-time
      isa operator
      pre other-not-inside
      action test-time
      arg1 shooter-timeout
      success shooter-waited-enough
      fail thrust-stop)

     (nothing
      isa operator
      pre nothing-to-do
      action restart
      post done)

     (shooter-change-role
      isa operator
      pre shooter-waited-enough
      action change-to-bait
      post done)

     (behind?
      isa operator
      pre other-inside
      action test-categorical-dimension
      arg1 location
      arg2 behindFortress
      success behind
      fail notbehind)

     (future-behind?
      isa operator
      pre notbehind
      action test-predicted-location-union
      arg1 behindFortress
      arg2 locationHorizon
      success behind
      fail wont-be-behind)

     (fortress-vulnerable?
      isa operator
      pre behind
      action test-categorical-dimension
      arg1 fortress-back
      arg2 vulnerable
      success vulnerable
      fail notvulnerable)

     (will-stay-behind?
      isa operator
      pre notvulnerable
      action test-predicted-location
      arg1 behindFortress
      arg2 shootHorizon
      success will-stay-behind
      fail thrust-stop)


     (aim-at-fortress-good?
      isa operator
      pre vulnerable
      action test-aim
      arg1 fortress-angle
      success shoot
      fail aimnotatfortress)

     (anticipatory-aim-at-fortress
      isa operator
      pre aimnotatfortress
      action pred-aim-at-fortress
      post shoot)

     (shoot-or-skip-shot
      isa operator
      pre shoot
      action tap-key
      arg1 shoot
      post done)

     (anticipatory-aim-at-fortress-no-shoot
      isa operator
      pre will-stay-behind
      action pred-aim-at-fortress
      post nothing-to-do)

     (predict-back-orientation
      isa operator
      pre wont-be-behind
      action pred-back-orientation
      post behind-direction)

     (determine-behind-direction
      isa operator
      pre behind-direction
      action calculate-behind-direction
      post check-flight-dir)

     (outside?-nothing
      isa operator
      pre nofortress
      action test-categorical-dimension
      arg1 location
      arg2 outsideHex
      success nothing-to-do
      fail notoutside)

     (stopped?
      isa operator
      pre thrust-stop
      action test-speed
      arg1 stopped
      success nothing-to-do
      slow aim-speed-up
      fast aim-slow-down)

     (future-outside?
      isa operator
      pre notoutside
      action test-predicted-location
      arg1 outsideHex
      arg2 aimHorizon
      fail outside-direction
      success aim-stop-outside)

     (anticipatory-aim-stop
      isa operator
      pre aim-stop-outside
      action aim-opposite
      post nothing-to-do)

     (set-target-speed-outside
      isa operator
      pre set-speed-stopped-outside
      action set-speed
      arg1 stopped
      post thrust-stop-outside)

     (thrust-stop-outside
      isa operator
      pre thrust-stop-outside
      action thrust-speed
      arg1 outsideHex
      post done)

     (determine-outside-direction
      isa operator
      pre outside-direction
      action calculate-outside-direction
      post check-flight-dir)

     (flight-direction-ok?
      isa operator
      pre check-flight-dir
      action test-flight-direction
      success check-speed
      fail thrust-angle)

     (speed-ok?
      isa operator
      pre check-speed
      action test-speed
      arg1 good-speed
      success nothing-to-do
      slow aim-speed-up
      fast aim-slow-down)

     (flight-direction-slow-ok?
      isa operator
      pre check-flight-dir-slow
      action test-flight-direction
      success check-speed-slow
      fail thrust-angle-slow)

     (speed-slow-ok?
      isa operator
      pre check-speed-slow
      action test-speed
      arg1 slow-speed
      success nothing-to-do
      slow aim-speed-up
      fast aim-slow-down)

     (aim-to-slow-down
      isa operator
      pre aim-slow-down
      action aim-opposite
      post change-speed)

     (aim-to-speed-up
      isa operator
      pre aim-speed-up
      action aim-flight-direction
      post change-speed)

     (thrust-for-speed
      isa operator
      pre change-speed
      action thrust-speed
      post done)

     ))

(defparameter *chunk-types*
  '((chunk-type gamestate x y orientation vx vy ship-alive status game  ;common
     offaim  moving-angle current-angle future-angle r1 r2 ship-speed   ;spacetrack
     time-to-outer thrust-angle vulnerability fortress-angle fortress-alive outward prop-to-big speed ;space fortess
     flight-direction location role other-x other-y other-vx other-vy other-location shell-x shell-y fortress-target ;coop
     fortress-back missile-present ;coop
     )   
    (chunk-type goal step state nextstep target aim game thrust-time ;common
     rectangle  offaim  press-point stop-angle ship-speed ;spacetrack
     time-to-outer time-thresh dist-thresh vulnerability last-action outward prop-to-big speed ;space fortess
     target-fd aim-at-fortress predicted-location target-speed shootHorizon locationHorizon shooter-timeout ;coop
     role shooter-distance bait-distance stopped slow-speed good-speed thrustHorizon tapaim fortress-back-angle ; coop
     aim-delta speed-delta start-fd aim-tolerance start-speed fortress-death ;coop
     )
    (chunk-type imaginal speed-reached speed-not-reached thrust-horizon-good thrust-horizon-bad good-shot skip-shot) ;coop
    (chunk-type operator pre action arg1 arg2 result post fail success slow fast)
    (chunk-type tracker rectangle death bad-aim good-aim
     hit miss badspeed goodspeed bighexdeath shelldeath smallhexdeath
     bad-weight
     )
    (chunk-type mapping function key)
    ))

(defparameter *default-chunks*
  '((default-chunks
     '(yes no bad speed adjust offaim turn-to-point test-dimension pressing-d pressing-key pressing-left down target-angle w a d counterclockwise-turn clockwise-turn fire  
       miss kill death ship-speed slow-speed stop-angle thrust-time 
       TAP-THRUST rectangle press-point tapthrust TAPPING-KEY pressing-a
       CALCULATE-TURN-ANGLE maketurn2 test-speed  start done play start-playing-again do-step game-over
       start-playing future-angle pressing-thrust-speed pressing-thrust-direction tap-key thrust-to-turn
       NOFORTRESS TESTOBJECT DELAY? CHECK-DELAY ANGLE fortress-angle  double-space-bar
       thrust thrust-angle increment-angle decrement-angle TIME-TO-OUTER space-bar shooting vulnerability? thrusting
       double-spacebar state-changed reset hit time-thresh bighexdeath spacebar test-object prop-to-big
       shelldeath smallhexdeath badspeed goodspeed
       shootHorizon insideHex outsideHex atFortress crashedIntoOuter shooter-timeout shooter-distance ; coop
       bait-distance stopped good-speed test-predicted-location predict-location pred-back-orientation ;coop
       calculate-outside-direction calculate-avoid-fortress-direction calculate-avoid-fortress-angle ;coop
       calculate-thrust-angle aim-opposite aim-flight-direction change-to-bait thrust-speed tap-key-loc ;coop
       calculate-behind-direction calculate-stay-inside-direction thrust-speed-loc shoot test-time test-categorical-dimension ;coop
       calculate-bait-direction atboundary crashed-at-outer wont-crash-outer thrust-opposite coop thrust-direction fortress-alive;coop
       no-crash-fortress crash-fortress yestarget notarget none fortress-target norole role test-role updaterole pick-role ;coop
       shooter bait bait-notinside bait-inside location bait-not-slow bait-slow bait-will-exit nothing-to-do ;coop
       bait-future-not-inside bait-future-inside thrust-slow-inside other-not-inside other-inside other-location ;coop
       shooter-waited-enough notbehind behind behindfortress wont-be-behind notvulnerable vulnerable fortress-back ;coop
       will-be-behind pred-aim-at-fortress thrust-stop behind-direction notoutside outside outside-direction ;coop
       thrust-angle-slow thrust-slow flight-direction-bad flight-direction-good test-bait-direction aim-stop-outside ;coop
       calculate-good-bait-direction restart evaluate-thrust-direction evaluate-thrust-speed ;coop
       evaluate-clockwise-turn evaluate-counterclockwise-turn aimnotatfortress test-aim predict-aim-at-fortress ;coop
       test-predicted-location-union crashedintofortress check-flight-dir-slow aim-slow-down
       aim-speed-up avoidexithorizon aimhorizon check-flight-dir set-speed
       set-speed-slow-inside will-stay-behind set-speed-stopped-outside check-speed
       test-flight-direction check-speed-slow change-speed locationhorizon
       ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; General productions for playing a SF-style game

(defparameter *productions*
  '(
  ;;; This production continues play at the very beginning
    (p get-started
     "Make sure game is running"
     ?game>
     game-state game
     ?goal>
     buffer empty
     ?imaginal>
     state free
     ?game-state>
     state free
     ==>
     +goal>
     isa goal
     state start-playing
     +imaginal>
     isa tracker
     +game-state>
     isa gamestate)

  ;;; This production continues play for another game.
    (p get-started-again
     "Make sure game is running"
     ?game>
     game-state game
     =game-state>
     ship-alive yes
     game =game
     =goal>
     state game-over
     game =game
     ==>
     =game-state>
     status nil
     =goal>
     state start-playing-again
     rectangle nil)

    (p get-started-new-game
     "Make sure game is running"
     ?game>
     game-state game
     ?game-state>
     state free
     =game-state>
     ship-alive yes
     game =game
     ?imaginal>
     state free
     =goal>
     state game-over
     - game =game
     ==>
     =goal>
     isa goal
     state start-playing
     +imaginal>
     isa tracker
     +game-state>
     isa gamestate)

    (p start-playing-coop
     "once the ship is visible get going"
     =goal>
     isa goal
     state start-playing
     -  game coop
     =game-state>
     game coop
     ?tracker>
     state free
     ==>
     +tracker>
     control-slot slow-speed
     good-slot shell-fired
     bad-slot death
     min 0
     max 0.4
     bad-weight -1
     +tracker>
     control-slot good-speed
     good-slot kill
     bad-slot death
     min 0.4
     max 2.6
     bad-weight -1
     +tracker>
     control-slot bait-distance
     good-slot kill
     bad-slot death
     min 0
     max 160
     bad-weight -1
     +tracker>
     control-slot shooter-distance
     good-slot kill
     bad-slot death
     min 0
     max 160
     bad-weight -1
     +tracker>
     control-slot avoidExitHorizon
     good-slot kill
     bad-slot death
     min 0
     max 50
     bad-weight -1
     +tracker>
     control-slot aimHorizon
     good-slot kill
     bad-slot death
     min 0
     max 100
     bad-weight -1
     +tracker>
     control-slot shootHorizon
     good-slot kill ;good-shot
     bad-slot miss ;skip-shot
     min 0
     max 100
     bad-weight -1
     +tracker>
     control-slot locationHorizon
     good-slot kill
     bad-slot death
     min 0
     max 120
     bad-weight -1
     +tracker>
     control-slot aim-tolerance
     good-slot kill
     bad-slot miss
     min 3
     max 45
     bad-weight -1
     =goal>
     state play
     game coop
     role none
     stopped 0
     shooter-timeout 35
     tapaim 28.8 
     aim 10.8 
     thrustHorizon 9
     speed-delta 0 
     aim-delta 0 
     )

    (p start-playing-again-coop
     =goal>
     isa goal
     state start-playing-again
     =game-state>
     ship-alive yes
     game coop
     ==>
     -retrieval>
     =goal>
     state play
     role none
     )

    (spp start-playing-coop :u 100)
    (spp start-playing-again-coop :u 100)

    (p detect-trial-end
     "If a trial is finished, clear things out"
     =goal>
     isa goal
     -  state game-over
     =game-state>
     status game-over
     ?manual>
     processor free
     ==>
     =goal>
     state game-over
     rectangle nil
     step nil
     +manual>
     isa release-all-fingers
     )
    (spp detect-trial-end :u 10000)

  ;;; This production handles death
    (p detect-destroyed
     =goal>
     isa goal
     - state start-playing-again
     =game-state>
     ship-alive no
     ?manual>
     processor free
     ==>
     -retrieval>
     =goal>
     state start-playing-again
     step nil
     +manual>
     isa release-all-fingers
     )

    (spp detect-destroyed :u 90 )

    (p do-something
     =goal>
     isa goal
     state play
     ?retrieval>
     state free
     ?MANUAL>
     state FREE
     =game-state>
     game =game
     ==>
     =goal>
     state do-step
     +retrieval>
     isa operator
     pre =game
     )

    (p do-something-fast
     =goal>
     isa goal
     state play
     last-action shooting
     ?retrieval>
     state free
     ?MANUAL>
     PROCESSOR FREE
     =game-state>
     game =game
     ==>
     =goal>
     state do-step
     last-action nil
     +retrieval>
     isa operator
     pre =game
     )

    (p object-present
     =goal>
     state do-step
     =retrieval>
     isa operator
     action test-object
     arg1 =object
     success =newstep
     =game-state>
     =object yes
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p object-not-present
     =goal>
     state do-step
     =retrieval>
     isa operator
     action test-object
     arg1 =object
     fail =newstep
     =game-state>
     =object no
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p aim-success
     =goal>
     state do-step
     aim-tolerance =tol
     =retrieval>
     isa operator
     action test-aim
     arg1 =arg1
     success =newstep
     =game-state>
     =arg1 =target
     orientation =current
     !safe-eval! (aim-good =target =current =tol)
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p aim-fail
     =goal>
     state do-step
     aim-tolerance =tol
     =retrieval>
     isa operator
     action test-aim
     arg1 =arg1
     fail =newstep
     =game-state>
     =arg1 =target
     orientation =current
     !safe-eval! (not (aim-good =target =current =tol))
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-dimension-success
     =goal>
     state do-step
     =dimension =thresh
     =retrieval>
     isa operator
     action test-dimension
     arg1 =dimension
     arg2 nil
     success =newstep
     =game-state>
     =dimension =val
     !safe-eval! (> =val =thresh)
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-dimension-fail
     =goal>
     state do-step
     =dimension =thresh
     =retrieval>
     isa operator
     action test-dimension
     arg1 =dimension
     arg2 nil
     fail =newstep
     =game-state>
     =dimension =val
     !safe-eval! (<= =val =thresh)
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-time-success
     =goal>
     state do-step
     =dimension =thresh
     =retrieval>
     isa operator
     action test-time
     arg1 =dimension
     success =newstep
     =temporal>
     ticks =val
     !safe-eval! (> =val =thresh)
     ==>
     -temporal>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-time-fail
     =goal>
     state do-step
     =dimension =thresh
     =retrieval>
     isa operator
     action test-time
     arg1 =dimension
     fail =newstep
     =temporal>
     ticks =val
     !safe-eval! (<= =val =thresh)
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-speed-success
     =goal>
     state do-step
     =category =current
     =retrieval>
     isa operator
     action test-speed
     arg1 =category
     success =newstep
     =game-state>
     speed =target
     !safe-eval! (speed-close =current =target)
     ==>
     +retrieval>
     isa operator
     pre =newstep)

    (p test-speed-slow
     =goal>
     state do-step
     =category =target
     =retrieval>
     isa operator
     action test-speed
     arg1 =category
     slow =newstep
     =game-state>
     speed =current
     !safe-eval! (and (not (speed-close =current =target))
		  (< =current =target))
     ==>
     =goal>
     target-speed =target
     +retrieval>
     isa operator
     pre =newstep)

    (p test-speed-fast
     =goal>
     state do-step
     =category =target
     =retrieval>
     isa operator
     action test-speed
     arg1 =category
     fast =newstep
     =game-state>
     speed =current
     !safe-eval! (and (not (speed-close =current =target))
		  (> =current =target))
     ==>
     =goal>
     target-speed =target
     +retrieval>
     isa operator
     pre =newstep)

    (p test-bait-direction-success
     =goal>
     state do-step
     =retrieval>
     isa operator
     action test-bait-direction
     success =newstep
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-eval! (flight-direction-bait-p =x =y =vx =vy)
     ==>
     +retrieval>
     isa operator
     pre =newstep
     )

    (p test-bait-direction-fail
     =goal>
     state do-step
     =retrieval>
     isa operator
     action test-bait-direction
     fail =newstep
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-eval! (not (flight-direction-bait-p =x =y =vx =vy))
     ==>
     +retrieval>
     isa operator
     pre =newstep
     )

    (p predict-location
     =goal>
     state do-step
     =horizon =param
     =retrieval>
     action test-predicted-location
     arg2 =horizon
     arg1 =target
     success =s
     fail =f
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     other-x =ox
     other-y =oy
     other-vx =ovx
     other-vy =ovy
     fortress-alive =fal
     !safe-bind! =pred-loc (predicted-location =x =y =vx =vy =ox =oy =ovx =ovy =fal =param)
     ==>
     =goal>
     state predict-location
     success =s
     fail =f
     target =target
     predicted-location =pred-loc
     )

    (p predict-location-union
     =goal>
     state do-step
     =horizon =param
     =retrieval>
     action test-predicted-location-union
     arg2 =horizon
     arg1 =target
     success =s
     fail =f
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     other-x =ox
     other-y =oy
     other-vx =ovx
     other-vy =ovy
     fortress-alive =fal
     !safe-bind! =pred-loc (predicted-location-union =x =y =vx =vy =ox =oy =ovx =ovy =fal =param)
     ==>
     =goal>
     state predict-location
     success =s
     fail =f
     target =target
     predicted-location =pred-loc
     )

    (p predicted-location-success
     =goal>
     state predict-location
     target =target-loc
     success =next
     predicted-location =pred-loc
     !safe-eval! (values-equal 'location =pred-loc =target-loc)
     ==>
     =goal>
     state do-step
     target nil
     success nil
     fail nil
     predicted-location nil
     +retrieval>
     isa operator
     pre =next
     )

    (p predicted-location-fail
     =goal>
     state predict-location
     target =target-loc
     fail =next
     predicted-location =pred-loc
     !safe-eval! (not (values-equal 'location =pred-loc =target-loc))
     ==>
     =goal>
     state do-step
     target nil
     success nil
     fail nil
     predicted-location nil
     +retrieval>
     isa operator
     pre =next
     )

    (p test-categorical-dimension-success
     =goal>
     state do-step
     =game-state>
     =dim =current
     =retrieval>
     action test-categorical-dimension
     arg1 =dim
     arg2 =target
     success =next
     !safe-eval! (values-equal =dim =current =target)
     ==> 
     +retrieval>
     isa operator
     pre =next
     )

    (p test-categorical-dimension-fail
     =goal>
     state do-step
     =game-state>
     =dim =current
     =retrieval>
     action test-categorical-dimension
     arg1 =dim
     arg2 =target
     fail =next
     !safe-eval! (not (values-equal =dim =current =target))
     ==> 
     +retrieval>
     isa operator
     pre =next
     )

    (p update-role-to-shooter
     =goal>
     state do-step
     role bait
     =game-state>
     fortress-target other
     =retrieval>
     isa operator
     action updaterole
     ==>
     =goal>
     role none
     +retrieval>
     isa operator
     pre norole)

    (p update-role-to-bait
     =goal>
     state do-step
     role shooter
     =game-state>
     fortress-target self
     =retrieval>
     isa operator
     action updaterole
     ==>
     -temporal>
     =goal>
     role none
     +retrieval>
     isa operator
     pre norole)

    (p set-role-to-shooter
     =goal>
     state do-step
     role none 
     =game-state>
     fortress-target other
     =retrieval>
     isa operator
     action updaterole
     post =next
     ==>
     =goal>
     role shooter
     +temporal>
     isa time
     +retrieval>
     isa operator
     pre =next)

    (p role-is-shooter
     =goal>
     state do-step
     role shooter
     =game-state>
     fortress-target other
     =retrieval>
     isa operator
     action updaterole
     post =next
     ==>
     +temporal>
     isa time
     +retrieval>
     isa operator
     pre =next)

    (p set-role-to-bait
     =goal>
     state do-step
     role none 
     =game-state>
     fortress-target self
     =retrieval>
     isa operator
     action updaterole
     post =next
     ==>
     =goal>
     role bait
     +temporal>
     isa time
     +retrieval>
     isa operator
     pre =next)

    (p role-is-bait
     =goal>
     state do-step
     role bait
     =game-state>
     fortress-target self
     =retrieval>
     isa operator
     action updaterole
     post =next
     ==>
     +temporal>
     isa time
     +retrieval>
     isa operator
     pre =next)

    (p role-present
     =goal>
     state do-step
     role =role
     =retrieval>
     isa operator
     action test-role
     arg1 nil
     success =next
     !safe-eval! (not (equal =role 'none))
     ==>
     +retrieval>
     isa operator
     pre =next)

    (p role-abscent
     =goal>
     state do-step
     role none 
     =retrieval>
     isa operator
     action test-role
     arg1 nil
     fail =next
     ==>
     +retrieval>
     isa operator
     pre =next)

    (p test-role-success
     =goal>
     state do-step
     role =role
     =retrieval>
     isa operator
     action test-role
     arg1 =role
     success =next
     ==>
     +retrieval>
     isa operator
     pre =next)

    (p test-role-fail
     =goal>
     state do-step
     role =role
     =retrieval>
     isa operator
     action test-role
     arg1 =arg1
     fail =next
     !safe-eval! (not (equalp =role =arg1))
     ==>
     +retrieval>
     isa operator
     pre =next)

    (p predict-back-orientation
     =goal>
     shootHorizon =horizon
     =retrieval>
     action pred-back-orientation
     post =next
     =game-state>
     other-x =ox
     other-y =oy
     other-vx =ovx
     other-vy =ovy
     !safe-bind! =fba (predict-back-orientation =ox =oy =ovx =ovy =horizon)
     ==>
     =goal>
     fortress-back-angle =fba
     +retrieval>
     isa operator
     pre =next
     )

    (p there-is-nothing-to-do
     =goal>
     state do-step
     =retrieval>
     isa operator
     action restart
     !eval! t
     ==>
     =goal>
     state play
     )

    (p calc-behind-direction
     =goal>
     shooter-distance =sd
     fortress-back-angle =fba
     locationHorizon =h
     =retrieval>
     action calculate-behind-direction
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-bind! =target-fd (behind-angle =x =y =vx =vy =fba =sd =h)
     ==>
     =goal>
     target-fd =target-fd
     fortress-back-angle nil
     +retrieval>
     isa operator
     pre =next
     )

    (p calc-good-bait-direction
     =goal>
     bait-distance =bd
     =retrieval>
     action calculate-good-bait-direction
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-bind! =target-fd (bait-inside-angle =x =y =vx =vy =bd)
     ==>
     =goal>
     target-fd =target-fd
     +retrieval>
     isa operator
     pre =next
     )

    (p calc-bait-direction
     =goal>
     bait-distance =bd
     =retrieval>
     action calculate-bait-direction
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     other-x =ox
     other-y =oy
     shell-x =sx
     shell-y =sy
     !safe-bind! =target-fd (bait-angle =x =y =vx =vy =ox =oy =sx =sy =bd)
     ==>
     =goal>
     target-fd =target-fd
     +retrieval>
     isa operator
     pre =next
     )

    (p calc-stay-inside-direction
     =goal>
     bait-distance =bd
     =retrieval>
     action calculate-stay-inside-direction
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-bind! =target-fd (stay-inside-angle =x =y =vx =vy =bd)
     ==>
     =goal>
     target-fd =target-fd
     +retrieval>
     isa operator
     pre =next
     )

    (p calc-outside-direction
     =goal>
     locationHorizon =horizon
     good-speed =gs
     =retrieval>
     action calculate-outside-direction
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-bind! =target-fd (outside-angle =x =y =vx =vy =horizon =gs)
     ==>
     =goal>
     target-fd =target-fd
     +retrieval>
     isa operator
     pre =next
     )
    
    (p calculate-avoid-fortress-angle
     =goal>
     state do-step
     =retrieval>
     isa operator
     action calculate-avoid-fortress-angle
     post =next
     =game-state>
     x =x
     y =y
     fortress-angle =fa
     !safe-bind! =target-fd (avoid-fortress-angle =x =y (angle-difference =fa 180))
     ==>
     =goal>
     target-fd =target-fd
     +retrieval>
     isa operator
     pre =next
     )

    (p predict-aim-at-fortress
     =goal>
     shootHorizon =horizon
     =retrieval>
     action pred-aim-at-fortress
     post =next
     =game-state>
     x =x
     y =y
     vx =vx
     vy =vy
     !safe-bind! =aim (predict-aim-at-fortress =x =y =vx =vy =horizon)
     ==>
     =goal>
     state predict-aim-at-fortress
     aim-at-fortress =aim
     nextstep =next
     )

    (p aim-at-fortress-predicted
     =goal>
     state predict-aim-at-fortress
     aim-at-fortress =aim
     ==>
     =goal>
     target =aim
     state turn-to-point
     aim-at-fortress nil
     )

    (p skip-aim-to-thrust
     =goal>
     state do-step
     =category =target-speed
     target-fd =target-fd
     =retrieval>
     isa operator
     action calculate-thrust-angle
     arg1 =category
     post =next
     =game-state>
     vx =vx
     vy =vy
     speed =speed
     flight-direction =current-fd
     !safe-bind! =tangle (compute-aim =target-fd =vx =vy =target-speed)
     !safe-eval! (and (angles-within =target-fd =current-fd 10)
		  (speed-close =speed =target-speed))
     !eval! t
     ==>
     =goal>
     state play
     target-fd nil)

    (p aim-to-thrust
     =goal>
     state do-step
     =category =target-speed
     target-fd =target-fd
     =retrieval>
     isa operator
     action calculate-thrust-angle
     arg1 =category
     post =next
     =game-state>
     vx =vx
     vy =vy
     speed =speed
     flight-direction =current-fd
     !safe-bind! =tangle (compute-aim =target-fd =vx =vy =target-speed)
     !safe-eval! (not (and (angles-within =target-fd =current-fd  10)
		       (speed-close =speed =target-speed)))
     !eval! t
     ==>
     =goal>
     nextstep =next
     state turn-to-point
     target =tangle)

    (p flight-direction-success
     =goal>
     state do-step
     target-fd =target-fd
     =retrieval>
     isa operator
     action test-flight-direction
     success =next
     =game-state>
     flight-direction =current-fd
     !safe-eval! (angles-within =target-fd =current-fd 10)
     ==>
     +retrieval>
     pre =next)

    (p flight-direction-fail
     =goal>
     state do-step
     target-fd =target-fd
     =retrieval>
     isa operator
     action test-flight-direction
     fail =next
     =game-state>
     flight-direction =current-fd
     !safe-eval! (not (angles-within =target-fd =current-fd 10))
     ==>
     +retrieval>
     pre =next)

    (p aim-opposite
     =goal>
     state do-step
     =retrieval>
     isa operator
     action aim-opposite
     post =next
     =game-state>
     flight-direction =fd
     !safe-bind! =tangle (angle-difference =fd 180)
     ==>
     =goal>
     nextstep =next
     state turn-to-point
     target =tangle
     target-fd =tangle)

    (p aim-flight-direction
     =goal>
     state do-step
     =retrieval>
     isa operator
     action aim-flight-direction
     post =next
     =game-state>
     flight-direction =fd
     ==>
     =goal>
     nextstep =next
     state turn-to-point
     target =fd)

    (p pick-bait
     =goal>
     state do-step
     role =role
     =game-state>
     fortress-target =ft
     =retrieval>
     isa operator
     action pick-role
     post =next
     !eval! t
     !safe-eval! (or
		  (equal =role 'bait)
		  (equal =ft 'self)
		  (and (equal =ft 'none) (equal =role 'none)))
     ==>
     =goal>
     role bait
     +retrieval>
     isa operator
     pre =next)

    (p pick-shooter
     =goal>
     state do-step
     role =role
     =game-state>
     fortress-target =ft
     =retrieval>
     isa operator
     action pick-role
     post =next
     !eval! t
     !safe-eval! (or
		  (equal =ft 'other)
		  (and (equal =ft 'none) (equal =role 'none)))
     ==>
     =goal>
     role shooter
     +retrieval>
     isa operator
     pre =next
     +temporal>
     isa time
     )

    (p switch-to-bait
     =goal>
     state do-step
     =retrieval>
     isa operator
     action change-to-bait
     post =next
     !eval! t
     ==>
     =goal>
     state play
     role bait
     )


    (p set-speed-p
     =goal>
     state do-step
     =arg1 =target
     =retrieval>
     isa operator
     action set-speed
     arg1 =arg1
     post =nextstep
     ==>
     =goal>
     target-speed =target
     +retrieval>
     isa operator
     pre =nextstep
     )
     
    (p skip-thrust-speed
     =goal>
     state do-step
     target-speed =target
     aim =aim-param
     =retrieval>
     isa operator
     action thrust-speed
     =game-state>
     speed =current 
     flight-direction =fd
     orientation =aim
     ?manual>
     state free
     !safe-eval! (or (speed-close =current =target)
                  (not
		    (or
		      (and (> =target =current) (angles-within =fd =aim =aim-param))
		      (and (< =target =current) (angles-within (angle-difference =fd 180) =aim =aim-param)))))
     !eval! t
     ==>
     =goal>
     state play)

    (p tap-thrust-speed
     =goal>
     state do-step
     target-speed =target
     aim =aim-param
     =retrieval>
     isa operator
     action thrust-speed
     arg1 nil
     =game-state>
     speed =current 
     flight-direction =fd
     orientation =aim
     ?manual>
     state free
     !safe-eval! (and
		  (not (speed-close =current =target)) (<= (abs (- =current =target)) 0.360)
		  (or
		   (and (> =target =current) (angles-within =fd =aim =aim-param))
		   (and (< =target =current) (angles-within (angle-difference =fd 180) =aim =aim-param))))
     ==>
     +manual>
     cmd delayed-punch
     finger middle
     hand left
     delay .09
     =goal>
     state play)

    (p thrust-change-speed
     =goal>
     state do-step
     target-speed =target
     aim =aim-param
     =retrieval>
     isa operator
     action thrust-speed
     arg1 nil
     =game-state>
     speed =current 
     flight-direction =fd
     orientation =aim
     ?manual>
     state free
     !safe-eval! (and (> (abs (- =current =target)) 0.360) 
		  (or
		   (and (> =target =current) (angles-within =fd =aim =aim-param))
		   (and (< =target =current) (angles-within (angle-difference =fd 180) =aim =aim-param))))
     ==>
     +manual>
     cmd hold-punch
     finger middle
     hand left
     =goal>
     target-fd nil
     start-fd =fd
     start-speed =current
     state pressing-thrust-speed)

    (p tap-thrust-change-speed-loc
     =goal>
     state do-step
     target-speed =target
     =retrieval>
     isa operator
     action thrust-speed
     arg1 =arg1
     =game-state>
     location =loc
     speed =current
     flight-direction =fd
     orientation =aim
     ?manual>
     state free
     !safe-eval! (values-equal 'location =loc =arg1)
     !safe-eval! (and
		  (not (speed-close =current =target)) (<= (abs (- =current =target)) 0.360))
     ==>
     +manual>
     cmd delayed-punch
     finger middle
     hand left
     delay .09
     =goal>
     state play)

    (p thrust-change-speed-loc
     =goal>
     state do-step
     target-speed =target
     =retrieval>
     isa operator
     action thrust-speed
     arg1 =arg1
     =game-state>
     location =loc
     speed =current
     flight-direction =fd
     orientation =aim
     ?manual>
     state free
     !safe-eval! (values-equal 'location =loc =arg1)
     !safe-eval! (> (abs (- =current =target)) 0.360) 
     ==>
     +manual>
     cmd hold-punch
     finger middle
     hand left
     =goal>
     start-fd =fd
     start-speed =current
     state pressing-thrust-speed)

    (p finish-thrust-speed
     =goal>
     state pressing-thrust-speed
     target-speed =target
     speed-delta =delta
     good-speed =gs
     =game-state>
     speed =current
     vx =vx
     vy =vy
     orientation =aim
     ?manual-left>
     middle down
     ?manual>
     state free
     !safe-eval! (if (acceleratingp =vx =vy =aim)
		     (or (>= (+ =current =delta) =target)
			 (>= (+ =current =delta) =gs))
		     (<= (- =current =delta) =target))
     ==>
     =goal>
     state evaluate-thrust-speed
     +manual>
     cmd release
     hand left
     finger middle)

    (p thrust-speed-evaluation
     =goal>
     state evaluate-thrust-speed
     target-speed =target
     start-speed =start
     start-fd =start-fd
     speed-delta =sd
     =game-state>
     speed =current
     flight-direction =current-fd
     ?manual>
     state free
     !safe-bind! =theta (if (angles-within =start-fd =current-fd 90)
			    (update-theta =sd (cond ((and (>= =target =start) (>= =current =target)) (- =target =current))
						  ((and (>= =target =start) (< =current =target)) (- =target =current))
						  ((and (< =target =start) (>= =current =target)) (- =current =target))
						  ((and (< =target =start) (< =current =target)) (- =current =target))))
			    (update-theta =sd (- (- =current) =target)))
     !eval! t
     ==>
     =goal>
     state play
     start-speed nil
     target-speed nil
     start-fd nil
     target-fd nil
     speed-delta =theta)

    (p tap-thrust-direction
     =goal>
     state do-step
     target-fd =target
     thrustHorizon =horizon
     =retrieval>
     isa operator
     action thrust-to-turn
     =game-state>
     flight-direction =current 
     vx =vx
     vy =vy
     speed =speed
     orientation =aim
     ?manual>
     state free
     !safe-bind! =predicted (predict-direction =vx =vy =aim =horizon)
     !safe-eval! (or (angles-within =target =current 15)
		  (if (> =speed 0.25) (direction-close =current =target =predicted)))
     ==>
     +manual>
     cmd delayed-punch
     finger middle
     hand left
     delay .09
     =goal>
     state play)

    (p thrust-direction
     =goal>
     state do-step
     target-fd =target
     thrustHorizon =horizon
     =retrieval>
     isa operator
     action thrust-to-turn
     =game-state>
     flight-direction =current 
     vx =vx
     vy =vy
     speed =speed
     orientation =aim
     ?manual>
     state free
     !safe-bind! =predicted (predict-direction =vx =vy =aim =horizon)
     !safe-eval! (and (not (angles-within =target =current 15))
		  (not (if (> =speed 0.25) (direction-close =current =target =predicted))))
     ==>
     +manual>
     cmd hold-punch
     finger middle
     hand left
     =goal>
     state pressing-thrust-direction
     start-fd =current)

    (p finish-thrust-direction
     =goal>
     state pressing-thrust-direction
     target-fd =target
     thrustHorizon =horizon
     good-speed =gs
     =game-state>
     flight-direction =current
     vx =vx
     vy =vy
     orientation =aim
     ?manual-left>
     middle down
     ?manual>
     preparation free
     !safe-bind! =predicted (predict-direction =vx =vy =aim =horizon)
     !safe-bind! =predicted-speed (predict-speed =vx =vy =aim =horizon)
     !safe-eval! (or (direction-close =current =target =predicted)
		  (>= =predicted-speed =gs))
     ==>
     =goal>
     ;; state evaluate-thrust-direction
     state play
     target-fd nil
     start-fd nil
     +manual>
     cmd release
     hand left
     finger middle)

    (p thrust-direction-evaluation
     =goal>
     state evaluate-thrust-direction
     target-fd =target
     thrustHorizon =th
     start-fd =start
     =game-state>
     flight-direction =current
     ?manual>
     state free
     ==>
     =goal>
     state play
     target-fd nil
     start-fd nil
     )

    (p tap-key
     =goal>
     state do-step
     =retrieval>
     action tap-key
     arg1 =function
     ==>
     +retrieval>
     isa mapping
     function =function
     =goal>
     state tap-key)

    (p tap-key-loc
     =goal>
     state do-step
     =retrieval>
     action tap-key-loc
     arg1 =function
     arg2 =target
     =game-state>
     location =current
     !safe-eval! (values-equal 'location =current =target)
     ==>
     +retrieval>
     isa mapping
     function =function
     =goal>
     state tap-key)

    (p press-spacebar-if-behind
     =goal>
     state tap-key
;     aim-tolerance =at
     =retrieval>
     key spacebar
     ?manual>
     state free
     =game-state>
;     aim-at-fortress =target
;     orientation =current
     missile-present no
    ; location behindFortress
     ;fortress-alive yes
;     x =x
;     y =y
;     vx =vx
;     vy =vy
;     orientation =aim
;     =imaginal>
;     ?imaginal>
;     state free
     ;!safe-eval! (or (angles-within =target =current =at)
     ;             (and (angles-within =target =current =at)
;		       (drifting-away =x =y =vx =vy =aim));)
     ==>
     =goal>
     state play
;     =imaginal>
;     good-shot 1 
     +manual>
     cmd delayed-punch
     finger index
     hand right
     delay .09)

;    (p skip-spacebar-if-aim-bad
;     =goal>
;     state tap-key
;     aim-tolerance =at
;     =retrieval>
;     key spacebar
;     =game-state>
;     aim-at-fortress =target
;     orientation =current
;     =imaginal>
;     ?imaginal>
;     state free
;     !safe-eval! (not (angles-within =target =current =at))
;     !eval! t
;     ==>
;     =goal>
;     state play
;     =imaginal>
;     skip-shot 1)

   ; (p skip-spacebar-if-not-behind
   ;  =goal>
   ;  state tap-key
   ;  =retrieval>
   ;  key spacebar
   ;  =game-state>
   ;  ;location =loc
   ;  fortress-alive =fa
   ;  =imaginal>
   ;  ?imaginal>
   ;  state free
   ;  !safe-eval! (or 
;		  (not (equal =loc 'behindFortress))
;		  (equal =fa 'no))
;     !eval! t
;;     ==>
;     =goal>
;     state play
;     =imaginal>
;     skip-shot 1)

    ;;productions for turning general to both games
    (p turn-to-point1
     =goal>
     state do-step
     =retrieval>
     isa operator
     action turn-to-point
     arg1 =slot
     post =next
     =game-state>
     =slot =angle
     ==>
     =goal>
     nextstep =next
     state turn-to-point
     TARGET =angle)

    (p skip-turn
     =goal>
     state turn-to-point
     target =angle2
     aim =aim
     nextstep =next
     step nil
     =game-state>
     orientation =angle1
     ?manual>
     state free
     !safe-eval! (< (angle-offset =angle2 =angle1) =aim)
     ==>
     +retrieval>
     isa operator
     pre =next
     =goal>
     state do-step
     nextstep nil)

    (p start-press-clockwise
     =goal>
     state turn-to-point
     target =angle2
     tapaim =tapaim
     step nil
     =game-state>
     orientation =angle1
     ?manual>
     state free
     !safe-eval! (clockwisep =angle2 =angle1)
     !safe-eval! (>= (angle-offset =angle2 =angle1) =tapaim)
     ==>
     +retrieval>
     isa mapping
     function clockwise
     =goal>
     step pressing-key)

    (p press-clockwise-tap
     =goal>
     state turn-to-point
     target =angle2
     aim =aim
     tapaim =tapaim
     step nil
     =game-state>
     orientation =angle1
     ?manual>
     state free
     !safe-eval! (clockwisep =angle2 =angle1)
     !safe-eval! (< (angle-offset =angle2 =angle1) =tapaim)
     !safe-eval! (>= (angle-offset =angle2 =angle1) =aim)
     ==>
     +retrieval>
     isa mapping
     function clockwise
     =goal>
     step tapping-key)

    (p start-press-counterclockwise
     =goal>
     state turn-to-point
     target =angle2
     tapaim =tapaim
     step nil
     =game-state>
     orientation =angle1
     ?manual>
     state free
     !safe-eval! (counterclockwisep =angle2 =angle1)
     !safe-eval! (>= (angle-offset =angle2 =angle1) =tapaim)
     ==>
     +retrieval>
     isa mapping
     function counterclockwise
     =goal>
     step pressing-key)

    (p press-counterclockwise-tap
     =goal>
     state turn-to-point
     target =angle2
     aim =aim
     tapaim =tapaim
     step nil
     =game-state>
     orientation =angle1
     ?manual>
     state free
     !safe-eval! (counterclockwisep =angle2 =angle1)
     !safe-eval! (< (angle-offset =angle2 =angle1) =tapaim)
     !safe-eval! (>= (angle-offset =angle2 =angle1) =aim)
     ==>
     +retrieval>
     isa mapping
     function counterclockwise
     =goal>
     step tapping-key)

    (p tapping-d
     =goal>
     step tapping-key
     nextstep =next
     =RETRIEVAL>
     key d
     ?manual>
     state free
     ==>
     =goal>
     state do-step
     step nil
     target nil
     nextstep nil
     +MANUAL>
     CMD DELAYED-PUNCH
     FINGER index
     HAND LEFT
     delay .09
     +retrieval>
     isa operator
     pre =next)

    (p pressing-d
     =goal>
     step pressing-key
     =RETRIEVAL>
     key d
     ?manual>
     state free
     ==>
     =goal>
     step pressing-d
     +MANUAL>
     CMD hold-punch
     FINGER index
     HAND LEFT)

    (p finish-press-d
     =goal>
     step pressing-d
     target =target
     aim-delta =ad
     =game-state>
     orientation =current
     ?manual-left>
     index down
     ?manual>
     preparation free
     !safe-eval! (<= (angle-difference2 (angle-difference =current =ad) =target) 0)
     ==>
     =goal>
     state evaluate-clockwise-turn
     step nil
     +manual>
     cmd release
     hand left
     finger index)

    (p clockwise-turn-evaluation
     =goal>
     state evaluate-clockwise-turn
     target =target
     nextstep =next
     aim-delta =ad
     =game-state>
     orientation =current
     ?manual>
     state free
     !safe-bind! =theta (update-theta =ad (angle-difference2 =current =target))
     ==>
     =goal>
     state do-step
     aim-delta =theta
     target nil
     nextstep nil
     +retrieval>
     isa operator
     pre =next)

    (p pressing-a
     =goal>
     step pressing-key
     =RETRIEVAL>
     key a
     ?manual>
     state free
     ==>
     =goal>
     step pressing-a
     +MANUAL>
     CMD hold-punch
     FINGER ring
     HAND LEFT)

    (p finish-press-a
     =goal>
     step pressing-a
     target =target
     aim-delta =ad
     =game-state>
     orientation =current
     ?manual-left>
     ring down
     ?manual>
     preparation free
     !safe-eval! (>= (angle-difference2 (angle-difference =current (- =ad)) =target) 0)
     ==>
     =goal>
     state evaluate-counterclockwise-turn
     step nil
     +manual>
     cmd release
     hand left
     finger ring)

    (p counterclockwise-turn-evaluation
     =goal>
     state evaluate-counterclockwise-turn
     target =target
     nextstep =next
     aim-delta =ad
     =game-state>
     orientation =current
     ?manual>
     state free
     !safe-bind! =theta (update-theta =ad (angle-difference2 =target =current))
     ==>
     =goal>
     state do-step
     aim-delta =theta
     target nil
     nextstep nil
     +retrieval>
     isa operator
     pre =next)

    (p tapping-a
     =goal>
     step tapping-key
     nextstep =next
     =RETRIEVAL>
     key a
     ?manual>
     state free
     ==>
     =goal>
     state do-step
     step nil
     target nil
     nextstep nil
     +MANUAL>
     CMD DELAYED-PUNCH
     FINGER ring
     HAND LEFT
     delay .09
     +retrieval>
     isa operator
     pre =next)

    (p detect-fortress-death
     =goal>
     fortress-death nil
     =game-state>
     fortress-alive no
     !eval! t
     ==>
     =goal>
     state play
     fortress-death yes
     step nil
     target nil
     nextstep nil
     target-speed nil
     target-fd nil
     +manual>
     isa release-all-fingers
     -retrieval>
     )

    (spp detect-fortress-death :u 100)

    (p detect-fortress-alive
     =goal>
     fortress-death yes
     =game-state>
     fortress-alive yes
     !eval! t
     ==>
     =goal>
     fortress-death nil
     role none)

    (spp detect-fortress-alive :u 100)

    (spp (update-role-to-shooter) :reward 5)
    (spp (switch-to-bait) :reward 9)
    (spp (detect-fortress-death) :reward 90)
    ))

(defun run-coop-experiment (&optional (speed 1) (draw t) (alpha1 0.1) (alpha2 0.1) (factor1 1) (factor2 1))
  (clear-all)
  (define-model-fct 'player1-model
      `((sgp :sf-player 0)
	(sgp :alpha ,alpha1
	     :initial-temp ,factor1)
	,@*sgps*
	,@*chunk-types*
	,@*default-chunks*
	(add-dm ,@*action-to-key-map*)
	(add-dm ,@*operators*)
	,@*productions*
	))
  (define-model-fct 'player2-model
      `((sgp :sf-player 1)
	(sgp :alpha ,alpha2
	     :initial-temp ,factor2)
	,@*sgps*
	,@*chunk-types*
	,@*default-chunks*
	(add-dm ,@*action-to-key-map*)
	(add-dm ,@*operators*)
	,@*productions*
	))
  (play-sf-games 40 :speed speed :draw draw :cont nil))

