'use strict';

const args = require('yargs')
      .option('id', {
          describe: "The subject Id"})
      .option('fullscreen', {
          describe: "Window starts in fullscreen mode",
          type: 'boolean',
          default: true})
      .option('vsync', {
          describe: "Wait for vsync on screen updates",
          type: 'boolean',
          default: true})
      .option('debug', {
          describe: 'Turn on debugging commands.',
          type: 'boolean',
          default: false})
      .option('host', {
          describe: "The server to connect to",
          default: 'lookup'})
      .option('port', {
          describe: "The port to connect on",
          default: 3000})
      .option('audio-record-device', {
          describe: "The device index to record from.",
          type: 'number',
          default: 0})
      .option('audio-playback-device', {
          describe: "The device index to play to.",
          type: 'number',
          default: 0})
      // .option('audio-freq', {
      //     describe: "Audio frequency at which to record in-game audio.",
      //     type: 'number',
      //     default: 48000})
      // .option('audio-format', {
      //     describe: "Byte format for audio recording.",
      //     choices: ['u8', 's16', 'f32'],
      //     default: 'f32'})
      .option('audio-samples', {
          describe: "Number of audio samples to collect at a time.",
          type: 'number',
          default: 2880})
      .option('audio-list-devices', {
          describe: "List audio devices and quit.",
          type: 'boolean',
          default: false})
      .option('audio-silence-threshold', {
          describe: "The silence threshold.",
          type: 'number',
          default: 20})
      .help()
      .argv;

// if (args.id === undefined) {
//     console.error('You must specify a unique identifier for this client with --id.');
//     process.exit(1);
// }

const path = require('path');

var resourceDir = '.';
// A hack to detect OSX + app context.
if (process.pkg) {
    var p = path.parse(process.argv[0]);
    var d = p.dir.split(path.sep);
    if (d.length >= 3 &&
        d[d.length-1] === 'MacOS' &&
        d[d.length-2] === 'Contents') {
        resourceDir = path.join(path.dirname(process.argv[0]), '../Resources/');
    }
}
console.log('Resource dir:', resourceDir);


const graphics = require('./graphics');
const TCPClient = require('./tcpclient').TCPClient;
const {MainMenu, InputLine, AudioTest, NetworkTest, Consent} = require('./mainmenu');

const https = require('https');

graphics.initGraphics(path.join(resourceDir, 'freesansbold.ttf'),
                      {'freq': 48000,
                       'format': 'f32',
                       'samples': args['audio-samples'],
                       'playbackDevice': args['audio-playback-device'],
                       'recordDevice': args['audio-record-device'],
                       'threshold': args['audio-silence-threshold']});

if (args['audio-list-devices']) {
    process.exit(0);
}

function lookupServer(callback) {
    https.get('https://andersonlab.net/perl/serverlocation.pl?action=get&pw=71b0',
              (res) => {
                  console.log('status', res.statusCode);
                  let raw = '';
                  res.on('data', (chunk) => { raw += chunk; });
                  res.on('end', () => {
                      var data = JSON.parse(raw);
                      if (data.success) {
                          callback(data.host, data.port);
                      } else {
                          throw new Error(`Error while locating server: ${data.reason}`);
                      }
                  });
              });
}

function startClient(subjectId, host, port) {
    console.log('Server:', host, port);
    // graphics.createWindows(args.fullscreen, 'SFCoop', args.vsync);
    graphics.setFullscreen(args.fullscreen);
    var client = new TCPClient(resourceDir, host, port, subjectId);
    client.debug = args.debug;
    client.init();
}

graphics.createWindows(false, 'SFCoop', args.vsync);

function startMenu(host, port) {
    var mm = new MainMenu(['Network Test',
                           'Audio Test',
                           'Participation Consent',
                           'Start Experiment',
                           'Quit']);
    mm.on('selection', (i) => {
        console.log('selection', i);
        if (i === 2) {
            var d = new Consent();
            d.on('done', () => { mm.start(); });
            d.start();
        } else if (i === 3) {
            var input = new InputLine('Enter Your Participant ID');
            input.text = args.id || '';
            input.on('abort', () => { mm.start(); });
            input.on('input', (txt) => {
                console.log('starting experiment', txt);
                startClient(txt, host, port);
            });
            input.start();
        } else if (i === 1) {
            var d = new AudioTest(resourceDir);
            d.on('done', () => { mm.start(); });
            d.start();
        } else if (i === 0) {
            var d = new NetworkTest(host, port);
            d.on('done', () => { mm.start(); });
            d.start();
        } else {
            console.log('quit');
        }
    });

    mm.start();
}

if (args.host === 'lookup') {
    lookupServer(startMenu);
} else {
    startMenu(args.host, args.port);
}
