#include "graphics.hh"

Menu::Menu(Window *w, std::vector<std::string>& sessions, std::vector<std::string>& conditions, int numScrambles, std::string &warning) {
  mWindow = w;
  mSessions = sessions;
  mConditions = conditions;
  mSubjectId = "";
  mMode = -1;
  mSession = -1;
  mCondition = -1;
  mScramble = -1;
  mNumScrambles = numScrambles;
  mScreenMode = 1;
  mDebug = false;
  mGraphics = 0;
  mScale = 0;
  mLineWidth = 1;
  mWarning = warning;

  mModes.push_back("Client");
  mModes.push_back("Manual Client");
  mModes.push_back("Server");
  // mModes.push_back("Hybrid");
  // mModes.push_back("Samemachine");

  for (size_t i=0; i<sessions.size(); i++) { mSessionRects.push_back(new SDL_Rect); }
  for (size_t i=0; i<conditions.size(); i++) { mConditionRects.push_back(new SDL_Rect); }
  for (int i=0; i<mNumScrambles; i++) { mScrambleRects.push_back(new SDL_Rect); }
  for (size_t i=0; i<mModes.size(); i++) { mModeRects.push_back(new SDL_Rect); }
}

Menu::~Menu() {
}

std::string Menu::modeName() {
  if (mMode < 0) return "";
  else return mModes[mMode];
}

bool Menu::canStart() {
  if (mMode == -1) return false;
  if (mMode >= 1 && mSubjectId.size() == 0) return false;
  if (mMode == 2 && mSessions.size() > 0 && mSession == -1) return false;

  return true;
}

bool Menu::inputActive() {
  return mMode >= 1;
}

bool Menu::sessionsActive() {
  return mMode == 2;
}

std::string Menu::inputText() {
  if( mMode == 1 ) return "Server Host Name";
  else return "Participants ID";
}

void Menu::updateRects() {
  mDebugRect.x = 20;
  mDebugRect.y = 20;
  mDebugRect.w = 200;
  mDebugRect.h = 50;

  mStartRect.w = 300;
  mStartRect.h = 50;
  mStartRect.x = mWindow->getWidth()/2 - mStartRect.w/2;
  mStartRect.y = mWindow->getHeight() - mStartRect.h - 20;

  // Screen Mode
  int y = 20, x = mWindow->getWidth() - 20;
  for( int i=0; i<3; i++ ) {
    mScreenModeRects[i].w = 200;
    mScreenModeRects[i].h = 50;
    mScreenModeRects[i].x = x - mScreenModeRects[i].w;
    mScreenModeRects[i].y = y;
    x -= mScreenModeRects[i].w + 5;
  }
  y += mScreenModeRects[0].h + 5;
  mOneToOneScaleRect.w = 200;
  mOneToOneScaleRect.h = 50;
  mOneToOneScaleRect.x = mWindow->getWidth() - mOneToOneScaleRect.w - 20;
  mOneToOneScaleRect.y = y;
  mTenTwentyFourScaleRect = mOneToOneScaleRect;
  mTenTwentyFourScaleRect.x -= mTenTwentyFourScaleRect.w + 5;
  mSevenTenScaleRect = mTenTwentyFourScaleRect;
  mSevenTenScaleRect.x -= mSevenTenScaleRect.w + 5;
  mSdlRect = mOneToOneScaleRect;
  mSdlRect.y += mSdlRect.h + 5;
  mCairoRect = mSdlRect;
  mCairoRect.x -= mCairoRect.w + 5;
  mLW1Rect = mSdlRect;
  mLW1Rect.w = 132;
  mLW1Rect.x = mWindow->getWidth() - mLW1Rect.w - 20;
  mLW1Rect.y += mLW1Rect.h + 5;
  mLW2Rect = mLW1Rect;
  mLW2Rect.x -= mLW2Rect.w + 5;
  mLW3Rect = mLW2Rect;
  mLW3Rect.x -= mLW3Rect.w + 5;

  // Modes
  {
    y = mWindow->getHeight()/2 - 200;
    int width = 0, height = 0;
    int pad = 20;
    for(size_t i=0; i<mModes.size(); i++) {
      double w, h;
      textSize( mWindow, mModes[i], 28, w, h );
      mModeRects[i]->w = std::max( int(w) + 10, 200 );
      mModeRects[i]->h = std::max( int(h) + 10, 50 );
      mModeRects[i]->y = y;
      width += mModeRects[i]->w;
      if (h > height) height = h;
    }

    width += pad * (mModes.size()-1);
    x = mWindow->getWidth()/2 - width/2;
    for(size_t i=0; i<mModes.size(); i++) {
      mModeRects[i]->x = x;
      x += mModeRects[i]->w + pad;
    }
  }
  // Sessions
  // y = mWindow->getHeight()/3 + 60;
  y = mWindow->getHeight()/2 + 60;
  for(size_t i=0; i<mSessions.size(); i++) {
    double w, h;
    textSize( mWindow, mSessions[i], 28, w, h );
    mSessionRects[i]->w = std::max( int(w) + 10, 200 );
    mSessionRects[i]->h = std::max( int(h) + 10, 50 );
    mSessionRects[i]->x = mWindow->getWidth()/2 - mSessionRects[i]->w/2;
    mSessionRects[i]->y = y;
    y += mSessionRects[i]->h + 5;
  }

  // Conditions
  y += 30;
  for(size_t i=0; i<mConditions.size(); i++) {
    double w, h;
    textSize( mWindow, mConditions[i], 28, w, h );
    mConditionRects[i]->w = std::max( int(w) + 10, 200 );
    mConditionRects[i]->h = std::max( int(h) + 10, 50 );
    mConditionRects[i]->x = mWindow->getWidth()/2 - mConditionRects[i]->w/2;
    mConditionRects[i]->y = y;
    y += mConditionRects[i]->h + 5;
  }

  // Scrambles
  int w = 50, h = 50, pad = 5;
  int cols = 6;
  x = mWindow->getWidth()/2 - mNumScrambles * (w+pad) / 2;
  y += 30;
  for(int i=0; i<mNumScrambles; i++) {
    mScrambleRects[i]->w = w;
    mScrambleRects[i]->h = h;
    mScrambleRects[i]->x = x + i%cols * (w + pad);
    mScrambleRects[i]->y = y + i/cols * (h + pad);
  }
}

void Menu::drawButton(const std::string& text, SDL_Rect *box, bool selected) {
  if( selected )
    cairo_set_source_rgb(mWindow->mCtx, 0, 1.0, 0.2);
  else
    cairo_set_source_rgb(mWindow->mCtx, 0.75, 0.75, 0.75);

  cairo_rectangle(mWindow->mCtx, box->x, box->y, box->w, box->h);
  cairo_fill(mWindow->mCtx);
  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  cairo_set_font_face( mWindow->mCtx, gCairoFont );
  cairo_set_font_size( mWindow->mCtx, 28 );

  drawText(mWindow->mCtx, text, box->x + box->w/2.0, box->y + box->h/2.0, 0, 0);
}

void Menu::redraw() {
  mWindow->createContext();
  updateRects();

  cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
  cairo_paint(mWindow->mCtx);

  SDL_Rect r;

  // version
  cairo_set_source_rgb( mWindow->mCtx, 0, 0, 0 );
  cairo_set_font_face( mWindow->mCtx, gCairoFont );
  cairo_set_font_size( mWindow->mCtx, 16 );
  drawText( mWindow->mCtx, gVersion, mWindow->getWidth()-5, mWindow->getHeight()-5, -1, -1 );

  // Input Area
  if( inputActive() ) {
    r.h = 50;
    r.w = mWindow->getWidth();
    r.x = 0;
    r.y = mWindow->getHeight()/2-25;
    cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
    cairo_rectangle( mWindow->mCtx, r.x, r.y, r.w, r.h );
    cairo_fill( mWindow->mCtx );
    cairo_set_font_face( mWindow->mCtx, gCairoFont );
    cairo_set_font_size( mWindow->mCtx, 28 );
    drawText( mWindow->mCtx, inputText(), mWindow->getWidth()/2, mWindow->getHeight()/2-30, 0, -1 );
    if( mSubjectId.size() > 0 ) {
      cairo_set_source_rgb( mWindow->mCtx, 0, 0, 0 );
      drawText( mWindow->mCtx, mSubjectId, mWindow->getWidth()/2, mWindow->getHeight()/2, 0, 0 );
    }
  }

  // Modes
  for(size_t i=0; i<mModeRects.size(); i++) {
    drawButton( mModes[i], mModeRects[i], i == mMode );
  }

  // Sessions
  if( sessionsActive() ) {
    for(size_t i=0; i<mSessionRects.size(); i++) {
      drawButton( mSessions[i], mSessionRects[i], i == mSession );
    }
  }
  // Conditions
  for(size_t i=0; i<mConditionRects.size(); i++) {
    drawButton( mConditions[i], mConditionRects[i], i == mCondition );
  }

  // Scrambles
  for(size_t i=0; i<mScrambleRects.size(); i++) {
    std::stringstream s;
    s << i;
    drawButton( s.str(), mScrambleRects[i], i == mScramble );
  }

  drawButton( "Windowed", &mScreenModeRects[0], mScreenMode == 0 );
  drawButton( "FullScreen", &mScreenModeRects[1], mScreenMode == 1 );
  drawButton( "Debug Mode", &mDebugRect, mDebug );
  if( canStart() )
    drawButton( "Start Experiment", &mStartRect, false );

  cairo_set_source_rgb( mWindow->mCtx, 1, 0.3, 0.3 );
  drawText( mWindow->mCtx, mWarning, mWindow->getWidth()/2, mStartRect.y - 50, 0, -1 );

  mWindow->destroyContext();
  SDL_RenderPresent( mWindow->mRenderer );
}

void Menu::mouseClick( const SDL_Event &e, bool& update, bool &done ) {
  SDL_Point p;
  p.x = e.button.x;
  p.y = e.button.y;
  if( SDL_PointInRect( &p, &mStartRect ) ) {
    if( canStart() ) {
      done = true;
    }
    return;
  }
  if( SDL_PointInRect( &p, &mDebugRect ) ) {
    mDebug = !mDebug;
    update = true;
    return;
  }
  if( SDL_PointInRect( &p, &mScreenModeRects[0] ) ) {
    mScreenMode = 0;
    mWindow->setFullscreen( false );
    update = true;
    return;
  }
  if( SDL_PointInRect( &p, &mScreenModeRects[1] ) ) {
    mScreenMode = 1;
    mWindow->setFullscreen( true );
    update = true;
    return;
  }
  // if( SDL_PointInRect( &p, &mOneToOneScaleRect )) {
  //   mScale = 0;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mTenTwentyFourScaleRect )) {
  //   mScale = 1;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mSevenTenScaleRect )) {
  //   mScale = 2;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mSdlRect )) {
  //   mGraphics = 0;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mCairoRect )) {
  //   mGraphics = 1;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mLW1Rect )) {
  //   mLineWidth = 1;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mLW2Rect )) {
  //   mLineWidth = 2;
  //   update = true;
  // }
  // if( SDL_PointInRect( &p, &mLW3Rect )) {
  //   mLineWidth = 3;
  //   update = true;
  // }

  for( size_t i=0; i<mModeRects.size(); i++) {
    if( SDL_PointInRect( &p, mModeRects[i] ) ) {
      mMode = i;
      update = true;
      return;
    }
  }

  if (sessionsActive()) {
    for( size_t i=0; i<mSessionRects.size(); i++) {
      if( SDL_PointInRect( &p, mSessionRects[i] ) ) {
        mSession = i;
        update = true;
        return;
      }
    }
  }

  for( size_t i=0; i<mConditionRects.size(); i++) {
    if( SDL_PointInRect( &p, mConditionRects[i] ) ) {
      mCondition = i;
      update = true;
      return;
    }
  }

  for( size_t i=0; i<mScrambleRects.size(); i++) {
    if( SDL_PointInRect( &p, mScrambleRects[i] ) ) {
      mScramble = i;
      update = true;
      return;
    }
  }
}

void Menu::run() {
  SDL_Event e;
  bool done = false;

  SDL_ShowCursor( SDL_ENABLE );
  redraw();

  while( !done ) {
    bool update = false;
    if ( SDL_WaitEvent( &e ) == 0 ) {
      printf( "SDL_WaitEvent SDL_Error: %s\n", SDL_GetError() );
      shutdown();
    }

    switch( e.type ) {
    case SDL_WINDOWEVENT:
      if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
        update = true;
      }
      break;
    case SDL_QUIT:
      shutdown();
      break;
    case SDL_KEYDOWN:
      if( e.key.keysym.sym == SDLK_q && e.key.keysym.mod & KMOD_ALT ) {
        shutdown();
      } else {
        if( inputActive() ) {
          if ((e.key.keysym.sym >= SDLK_0 && e.key.keysym.sym <= SDLK_9) ||
              e.key.keysym.sym == SDLK_PERIOD || e.key.keysym.sym == SDLK_KP_PERIOD ||
              (e.key.keysym.sym >= SDLK_KP_0 && e.key.keysym.sym <= SDLK_KP_9) ||
              (e.key.keysym.sym >= SDLK_a && e.key.keysym.sym <= SDLK_z)) {
            std::string letter( SDL_GetKeyName( e.key.keysym.sym ) );
            if( !( e.key.keysym.mod & KMOD_SHIFT )) {
              std::transform( letter.begin(), letter.end(), letter.begin(), ::tolower );
            }
            mSubjectId.append( letter );
            update = true;
          } else if ( e.key.keysym.sym == SDLK_BACKSPACE ) {
            if (mSubjectId.size() > 0)
              mSubjectId.erase(mSubjectId.end()-1);
            update = true;
          }
        }
      }
      break;
    case SDL_MOUSEBUTTONDOWN:
      mouseClick( e, update, done );
      break;
    }

    if( update ) redraw();
  }
  SDL_ShowCursor( SDL_DISABLE );
}
