#include "graphics.hh"

const int BROADCAST_PORT = 2999;
const Uint8 BROADCAST_MAGIC = 0xAA;
const Uint32 MESSAGE_MAGIC = 0xDEADBEEF;

const Uint8 MESSAGE_TYPE_STRING = 0;
const Uint8 MESSAGE_TYPE_AUDIO = 1;
const Uint8 MESSAGE_TYPE_BUFFER = 2;
const Uint8 MESSAGE_TYPE_SPY_AUDIO_0 = 3;
const Uint8 MESSAGE_TYPE_SPY_AUDIO_1 = 4;

typedef struct {
  Uint32 magic = 0x12345678;
  Uint16 type = 0xFFFF;
  Uint32 checksum = 0;
  Uint32 size = 0;
} MessageHeader;


class Server {
public:
  Uint16 mPort;
  IPaddress mIP;
  TCPsocket mSocket;
  UDPpacket *mBroadcastPacket;
  UDPsocket mBroadcastSocket;
  SDLNet_SocketSet mSet;
  std::vector<TCPsocket>mClients;

  bool mConnected;

  Server(int port);
  ~Server();

  bool start();
  bool broadcastAvailability();
  bool checkForNewClients();
  void shutdown();
  void getClientMessages(size_t clientIndex, std::vector<Uint8> &types, std::vector<std::string> &messages);
  bool sendMessage(std::string &msg);
  bool sendMessage(std::string &msg, Uint8 player);
  bool sendAudioData(const std::string &data, Uint8 player);
};

class BroadcastClient {
public:
  UDPpacket *mPacket;
  UDPsocket mSocket;
  IPaddress mLastServer;

  BroadcastClient();
  ~BroadcastClient();

  bool checkForServerBroadcasts();
};

class Client {
public:
  std::string mHost;
  Uint16 mPort;
  IPaddress mIP;
  TCPsocket mSocket;
  SDLNet_SocketSet mSet;
  bool mConnected;

  Client(const std::string &host, int port);
  ~Client();

  bool connect();
  void getServerMessages(std::vector<Uint8> &types, std::vector<std::string> &messages);

  bool sendMessage(std::string &msg);
  bool sendAudioData(Uint8 *data, int bytes);
  bool sendData(Uint8 type, const void *data, Uint32 bytes);
};
