const path = require('path');
const graphics = require('./graphics');

function GraphicsHelper(resourceDir) {
    console.log('graphics', 'resourcedir', resourceDir);
    this.resourceDir = resourceDir;
    this.drawTimes = 0;
    this.drawTimeStamp = new Date().getTime();
    this.fps = 0;
    for (let i=0; i<this.drawTimes.length; i++)
        this.drawTimes[i] = 0;
}

GraphicsHelper.prototype = {};


GraphicsHelper.prototype.drawSpyScreenState = function(banner, data) {
    for (let i=0; i<graphics.numWindows(); i++) {
        // console.log('drawing', i, data);
        graphics.startDrawing(i);
        if (data.screen === 'message') this.drawMessage(i, data);
        else if (data.screen === 'game') this.drawGame(i, data);
        else if (data.screen === 'chat') this.drawChat(i, data);
        else if (data.screen === 'alert') this.drawAlert(i, data);
        else if (data.screen === 'wrappedMessage') this.drawWrappedMessage(i, data);

        this.drawBanner(i, banner);

        graphics.finishDrawing(i);
    }
    for (let i=0; i<graphics.numWindows(); i++) {
        // console.log('flip', i);
        graphics.flip(i);
    }
    this.drawTimes += 1;
    if (this.drawTimes % 30 === 0) {
        var now = new Date().getTime();
        this.fps = 1000 * 30 / (now - this.drawTimeStamp);
        this.drawTimeStamp = now;
    }
}

GraphicsHelper.prototype.drawScreenState = function(data) {
    for (let i=0; i<graphics.numWindows(); i++) {
        // console.log('drawing', i, data);
        graphics.startDrawing(i);
        if (data.screen === 'message') this.drawMessage(i, data);
        else if (data.screen === 'game') this.drawGame(i, data);
        else if (data.screen === 'chat') this.drawChat(i, data);
        else if (data.screen === 'alert') this.drawAlert(i, data);
        else if (data.screen === 'wrappedMessage') this.drawWrappedMessage(i, data);

        graphics.finishDrawing(i);
    }
    for (let i=0; i<graphics.numWindows(); i++) {
        // console.log('flip', i);
        graphics.flip(i);
    }
    this.drawTimes += 1;
    if (this.drawTimes % 30 === 0) {
        var now = new Date().getTime();
        this.fps = 1000 * 30 / (now - this.drawTimeStamp);
        this.drawTimeStamp = now;
    }
};

GraphicsHelper.prototype.splitLines = function (w, input, maxWidth) {
    var words = input.split(' ');
    var normalFontSize = 18;
    var fsize = graphics.fontSize(w, normalFontSize);
    var lines = [];
    var line = words.length>0?words[0]:'';

    for (let i=1; i<words.length; i++) {
        var next = line + ' ' + words[i];
        var s  = graphics.textSize(w, next, normalFontSize);
        if (s.w > maxWidth) {
            lines.push(line);
            line = words[i];
        } else {
            line = next;
        }
    }
    lines.push(line);

    return lines;
}

GraphicsHelper.prototype.drawBanner = function(w, data) {
    var size = graphics.screenSize(w);
    var s = [graphics.textSize(w, data[0], 18),
             graphics.textSize(w, data[1], 18)];
    this.drawRect(w, 0, 0, size.w, 25, 0xFFCCFF, true);
    this.drawRect(w, 0, size.h-25, size.w, 25, 0xFFCCFF, true);
    graphics.drawText(w, data[0], size.w/2-s[0].w/2, 18, 18, 0x000000);
    graphics.drawText(w, data[1], size.w/2-s[1].w/2, size.h-7, 18, 0x000000);
}

GraphicsHelper.prototype.drawMessage = function(w, data) {
    // var bg = 0x000000;
    // var fg = 0xFFFFFF;
    // var hl = 0xFFFF00;

    var bg = data.bg === undefined ? 0xE6F0FF:data.bg;
    var fg = data.fg === undefined ? 0x000000:data.fg;
    var hl = data.hl === undefined ? 0x999999:data.hl;

    graphics.clearScreen(w, bg);
    var size = graphics.screenSize(w);

    // this.w.render.color = 0x000000;
    // this.w.render.clear();
    // this.w.render.color = 0xFFFF00;

    // console.log(surface.SDL_Surface);

    var normalFontSize = 18;
    var titleFontSize = 24;
    var fsize = graphics.fontSize(w, normalFontSize);
    var titlesize = graphics.fontSize(w, titleFontSize);
    var lines = new Array(data.lines.length)
    var maxx = 0;
    var height = 0;
    var float_maxx = 0;
    var float_height = 0;
    for (let i=0; i<data.lines.length; i++) {
        if (typeof data.lines[i] === 'object') {
            if (data.lines[i].image) {
                lines[i] = graphics.imageSize(path.join(this.resourceDir, ...data.lines[i].image));
                // console.log(lines[i]);
                if (data.lines[i].float === 'right') {
                    float_height = height + lines[i].h;
                    if (lines[i].w > float_maxx) float_maxx = lines[i].w;
                } else {
                    height += lines[i].h;
                    if (lines[i].w > maxx) maxx = lines[i].w;
                }
            } else if (data.lines[i].title) {
                lines[i] = graphics.textSize(w, data.lines[i].title, titleFontSize);
                height += titlesize.height;
                if (lines[i].w > maxx) maxx = lines[i].w;
            } else {
                lines[i] = {w:0,h:0};
            }
        } else {
            lines[i] = graphics.textSize(w, data.lines[i], normalFontSize);
            height += fsize.height;
            if (lines[i].w > maxx) maxx = lines[i].w;
        }
    }
    if (float_height > height) height = float_height;
    // console.log(this.w.size);
    var y = (size.h - height)/2;
    var x = (size.w - (maxx+float_maxx))/2;

    for (let i=0; i<lines.length; i++) {
        if (typeof data.lines[i] === 'object') {
            if (data.lines[i].image) {
                if (data.lines[i].float === 'right') {
                    graphics.drawImage(w, path.join(this.resourceDir, ...data.lines[i].image), size.w/2-(maxx+float_maxx)/2+maxx, y, 'left', 'top');
                } else if (data.lines[i].x === 'left') {
                    graphics.drawImage(w, path.join(this.resourceDir, ...data.lines[i].image), x, y-fsize.ascent, 'left', 'top');
                    y += lines[i].h;
                } else {
                    graphics.drawImage(w, path.join(this.resourceDir, ...data.lines[i].image), size.w/2, y-fsize.ascent, 'center', 'top');
                    y += lines[i].h;
                }
            } else if (data.lines[i].title) {
                var fx;
                if (data.lines[i].float === 'center') fx = (size.w-lines[i].w)/2;
                else fx = x;
                graphics.drawText(w, data.lines[i].title, fx, y, titleFontSize, fg);
                graphics.drawPolygon(w, [{x:x, y:y+fsize.descent+2},{x:x+maxx,y:y+fsize.descent+2}], fg);
                y += fsize.height;
            }
        } else {
            graphics.drawText(w, data.lines[i], x, y, normalFontSize, fg);
            y += fsize.height;
        }
    }

    if (Array.isArray(data.continue)) {
        var cont = new Array(data.continue.length);
        for (let i=0; i<data.continue.length; i++) {
            cont[i] = graphics.textSize(w, data.continue[i], 18);
        }
        graphics.drawText(w,
                          data.continue[0],
                          size.w/2-cont[0].w-50,
                          size.h-cont[0].h-20,
                          18, hl);
        graphics.drawText(w,
                          data.continue[1],
                          size.w/2+50,
                          size.h-cont[0].h-20,
                          18, hl);
    } else {
        var ts = graphics.textSize(w, data.continue, 18);
        graphics.drawText(w,
                          data.continue,
                          size.w/2-ts.w/2,
                          size.h-ts.h-20,
                          18, hl);
    }

    if (data.corner !== undefined) {
        var tz = graphics.textSize(w, data.corner, 14);
        graphics.drawText(w, data.corner, 5, tz.h+5, 14, hl);
    }
};

GraphicsHelper.prototype.drawWrappedMessage = function (w, data) {
    // var bg = 0x000000;
    // var fg = 0xFFFFFF;
    // var hl = 0xFFFF00;

    var bg = 0xE6F0FF;
    var fg = 0x000000;
    var hl = 0x999999;

    graphics.clearScreen(w, bg);
    var size = graphics.screenSize(w);

    // this.w.render.color = 0x000000;
    // this.w.render.clear();
    // this.w.render.color = 0xFFFF00;

    // console.log(surface.SDL_Surface);

    var normalFontSize = 18;
    var titleFontSize = 24;
    var fsize = graphics.fontSize(w, normalFontSize);
    var titlesize = graphics.fontSize(w, titleFontSize);

    var maxx = 700;

    var result = [];
    for (let i=0; i<data.lines.length; i++) {
        if (typeof data.lines[i] === 'object') {
            result.push(data.lines[i]);
        } else {
            var r = this.splitLines(w, data.lines[i], maxx);
            // console.log('split', r);
            result.push(...r);
        }
    }

    var lineSizes = new Array(result.length)
    var height = 0;
    var float_height = 0;
    for (let i=0; i<result.length; i++) {
        if (typeof result[i] === 'object') {
            if (result[i].title) {
                lineSizes[i] = graphics.textSize(w, result[i].title, titleFontSize);
                height += titlesize.height;
                if (lineSizes[i].w > maxx) maxx = lineSizes[i].w;
            } else {
                lineSizes[i] = {w:0,h:0};
            }
        } else {
            lineSizes[i] = graphics.textSize(w, result[i], normalFontSize);
            height += fsize.height;
            if (lineSizes[i].w > maxx) maxx = lineSizes[i].w;
        }
    }
    if (float_height > height) height = float_height;
    // console.log(this.w.size);
    //var y = (size.h - height)/2;
    var y = 100;
    var x = (size.w - maxx)/2;

    for (let i=0; i<lineSizes.length; i++) {
        if (typeof result[i] === 'object') {
            if (result[i].title) {
                var fx;
                if (result[i].float === 'center') fx = (size.w-lineSizes[i].w)/2;
                else fx = x;
                graphics.drawText(w, result[i].title, fx, y, titleFontSize, fg);
                graphics.drawPolygon(w, [{x:x, y:y+fsize.descent+2},{x:x+maxx,y:y+fsize.descent+2}], fg);
                y += fsize.height;
            }
        } else {
            graphics.drawText(w, result[i], x, y, normalFontSize, fg);
            y += fsize.height;
        }
    }

    var ts = graphics.textSize(w, data.continue, 18);
    graphics.drawText(w,
                      data.continue,
                      size.w/2-ts.w/2,
                      size.h-ts.h-20,
                      18, hl);
}


// Connection.prototype.drawExplosion = function(x, y, frame) {
//     this.w.render.copy(this.explosion.texture(this.w.render), [frame * 96,0,96,96],
//                        [this.center.x + x-96/2,
//                         this.center.y - y-96/2,
//                         96,96]);
// }

GraphicsHelper.prototype.drawDebug = function(w, data) {
    var y = 20;
    graphics.drawText(w, 'FPS: ' + this.fps.toFixed(0), 10, y, 18, 0x999999);
    y += 20;
    if ('sendDelay' in data) {
        graphics.drawText(w, 'Delay: ' + data.sendDelay, 10, y, 18, 0x999999);
        y += 20;
    }
    if (data.latency) {
        for (let i=0; i<data.latency.length; i++) {
            graphics.drawText(w, 'P'+(i+1)+' Ping: ' + data.latency[i].toString() + 'ms', 10, y, 18, 0x999999);
            y += 20;
        }
    }
};

GraphicsHelper.prototype.drawAlertPopUp = function(w, lines) {
    size = graphics.screenSize(w);
    this.drawRect(w, 100, size.h/2 - 100,
                  size.w-200, 200,
                  0xFF0000, true);
    this.drawRect(w, 100+5, size.h/2 - 100+5,
                  size.w-200-10, 200-10,
                  0x666666, true);
    var fsize = graphics.fontSize(w, 18);
    var height = fsize.height * lines.length;
    var y = size.h/2 - height/2 + fsize.height/2;
    for (let i=0; i<lines.length; i++) {
        var ts = graphics.textSize(w, lines[i], 18);
        graphics.drawText(w, lines[i], size.w/2 - ts.w/2, y, 18, 0xFFFFFF);
        y += fsize.height;
    }
}

GraphicsHelper.prototype.drawGame = function(w, data, debug) {
    var size = graphics.screenSize(w);

    graphics.clearScreen(w, 0x000000);
    if (data.debug) this.drawDebug(w, data);
    graphics.drawFlatScore(w, data.points);
    graphics.setupGameArea(w);
    // this.w.render.color = 0x00ff00;
    // this.w.render.drawRect([[this.center.x-350,this.center.y-310,710,630]]);

    graphics.drawGameArea(w);

    graphics.fillHexagon(w, data.bighex.radius, data.bighex.x, data.bighex.y, 0, true, 0x222222);

    if (data.fortress.alive) {
        graphics.drawHexagon(w, data.smallhex.radius, data.smallhex.x, data.smallhex.y, data.smallhex.angle, data.smallhex.full, 0x00ff00);
    }

    for (let i=0; i<data.ships.length; i++) {
        if (data.ships[i].alive) {
            graphics.drawWireFrame(w, 'ship', data.ships[i].x, data.ships[i].y, data.ships[i].angle, data.ships[i].color);
        } else {
            graphics.drawExplosion(w, data.ships[i].x, data.ships[i].y);
        }
    }

    if (data.fortress.alive) {
        graphics.drawWireFrame(w, 'fortress', data.fortress.x, data.fortress.y, data.fortress.angle, 0xFFFF00);
    } else if (data.fortress.explosion) {
        graphics.drawExplosion(w, data.fortress.x, data.fortress.y);
    }

    for (let i=0; i<data.missiles.length; i++) {
        graphics.drawWireFrame(w, 'missile', data.missiles[i].x, data.missiles[i].y, data.missiles[i].angle, 0xFFFFFF);
    }

    for (let i=0; i<data.shells.length; i++) {
        graphics.drawWireFrame(w, 'shell', data.shells[i].x, data.shells[i].y, data.shells[i].angle, 0xFF0000);
    }

    // graphics.drawPolygon(w, data.fortress.shieldPoints, 0xff00ff);
    // graphics.drawPolygon(w, data.fortress.box, 0xff00ff);

    graphics.takeDownGameArea(w);
    if (data.disconnected) {
        this.drawAlertPopUp(w, ["Your partner's connection has been lost.",
                                "Please wait."]);
    }
};

GraphicsHelper.prototype.drawAlert = function (w, data) {
    graphics.clearScreen(w, 0xE6F0FF);
    this.drawAlertPopUp(w, data.lines);
}

GraphicsHelper.prototype.drawRect = function (w, x, y, width, height, color, fill) {
    graphics.drawPolygon(w, [{x:x, y:y},
                             {x:x+width, y:y},
                             {x:x+width, y: y+height},
                             {x:x, y: y+height},
                             {x:x, y:y}],
                         color, fill);
};

GraphicsHelper.prototype.drawChat = function (w, data) {
    var _this = this;
    var bg = data.bg === undefined ? 0xE6F0FF:data.bg;
    var fg = data.fg === undefined ? 0x000000:data.fg;
    var inputColor = 0xBFBFBF;
    var gray = 0x999999;
    var darkGray = 0x939393;
    var playerColors = [0x9999FF, 0xAAAAAA];

    graphics.clearScreen(w, bg);
    var size = graphics.screenSize(w);

    var normalFontSize = 18;
    var fsize = graphics.fontSize(w, normalFontSize);

    var pad = 40;
    var inputPad = 5;
    var maxBubbleWidth = 400;
    var historyWidth = 700;
    var historyHeight = 500;
    var rowHeight = fsize.height + inputPad * 4;
    var startY = (size.h - (historyHeight + rowHeight * 2))/2;

    var messageTotalHeight = 0;
    var wrappedHistoryText = new Array(data.history.length);
    for (let i=0; i<data.history.length; i++) {
        wrappedHistoryText[i] = this.splitLines(w, data.history[i].text, maxBubbleWidth);
        messageTotalHeight += wrappedHistoryText[i].length * fsize.height + 9;
    }
    if (data.otherPlayerStartedTyping) messageTotalHeight += fsize.height + 9;

    var columnStartX = function (column, textWidth) {
        var x;
        if (column) x = (size.w - historyWidth)/2 + inputPad * 2 - 3;
        else x = size.w/2 + historyWidth/2 - inputPad*2 - textWidth - 7;
        return x;
    };

    var bubble = function (y, lines, column) {
        var bubbleColor = playerColors[column];
        var tz = new Array(lines.length);
        var longest = 0;
        for (var i=0; i<lines.length; i++) {
            tz[i] = graphics.textSize(w, lines[i], normalFontSize);
            if (longest < tz[i].w) longest = tz[i].w;
        }
        var x = columnStartX(column, longest);
        var startY = y - lines.length * fsize.height;
        y = startY;
        graphics.fillBubble(w, x, y-fsize.height+3, longest + inputPad*2, fsize.height*lines.length+5, 5, bubbleColor);
        for (var i=0; i<lines.length; i++) {
            graphics.drawText(w, lines[i], x+inputPad, y, normalFontSize, fg);
            y += fsize.height;
        }
        // return y + 9;
        return startY-9;
    };

    this.drawRect(w, (size.w-historyWidth)/2, startY, historyWidth, historyHeight+rowHeight*2, fg);
    graphics.drawPolygon(w, [{x:(size.w-historyWidth)/2, y:startY+historyHeight},
                             {x:size.w/2+historyWidth/2, y:startY+historyHeight}],
                         fg);
    graphics.drawPolygon(w, [{x:(size.w-historyWidth)/2, y:startY+historyHeight+rowHeight},
                             {x:size.w/2+historyWidth/2, y:startY+historyHeight+rowHeight}],
                         fg);

    graphics.clipDrawingArea(w, (size.w-historyWidth)/2, startY+2, historyWidth, historyHeight-4);

    var hasTyped = false;
    var y;
    if (messageTotalHeight > historyHeight)
        y = startY + historyHeight + 9;
    else
        y = startY + messageTotalHeight + fsize.height - 5;

    if (data.otherPlayerTyping) {
        bubble(y, ['...'], 1);
        y -= fsize.height + 9;
    } else if (data.otherPlayerStartedTyping) {
        y -= fsize.height + 9;
    }
    // var startIdx = data.history.length - 14;
    // if (startIdx < 0) startIdx = 0;
    for (let i = data.history.length-1; i>=0; i--) {
        // graphics.fillBubble(w, x-inputPad, y-tz.h-3, tz.w + inputPad*2, tz.h+8, 5, playerColors[data.history[i].column]);
        // graphics.drawText(w, data.history[i].text, x, y, normalFontSize, fg);
        if (data.history[i].type === 'status') {
            var tz = graphics.textSize(w, wrappedHistoryText[i][0], normalFontSize);
            var x = columnStartX(data.history[i].column, tz.w);
            graphics.drawText(w, wrappedHistoryText[i][0], x, y-fsize.height, normalFontSize, gray);
            y -= fsize.height + 9;
        } else {
            y = bubble(y, wrappedHistoryText[i], data.history[i].column);
        }

        if (data.history[i].column === 0 && data.history[i].type !== 'status') hasTyped = true;
    }

    graphics.unclipDrawingArea(w);

    // this.drawRect(w, (size.w-historyWidth)/2, size.h-pad*2-inputPad, historyWidth, fsize.height+inputPad*2, fg);

    var helpText = '';
    var tz = graphics.textSize(w, data.text, normalFontSize);
    var inputWidth = historyWidth - inputPad*2;
    var inputStartX;

    if (tz.w < inputWidth - inputPad - 2) {
        inputStartX = (size.w-historyWidth)/2+inputPad*2;
    } else {
        inputStartX = (size.w-historyWidth)/2+inputPad*2 + historyWidth - inputPad*4 - tz.w;
    }

    this.drawRect(w, (size.w-historyWidth)/2 + inputPad, startY+historyHeight+8, inputWidth, fsize.height+4, inputColor, data.focus === 'text');

    graphics.clipDrawingArea(w, (size.w-historyWidth)/2 + inputPad, startY+historyHeight+8, inputWidth, fsize.height+4);

    if (data.text.length > 0) {
        graphics.drawText(w, data.text, inputStartX, startY + historyHeight+8+4 + fsize.height*2/3, normalFontSize, data.focus === 'text' ? fg:gray);
    } else if (data.focus === 'text' && !hasTyped) {
        graphics.drawText(w, "Type your message and press Enter", (size.w-historyWidth)/2+inputPad*2+5, startY + historyHeight+8+4 + fsize.height*2/3, normalFontSize, darkGray);
    }

    graphics.unclipDrawingArea(w);

    if (data.focus === 'text') {
        helpText = 'To select the "Continue" option, press Down.';
        graphics.drawPolygon(w, [{x: inputStartX+tz.w + 2, y: startY + historyHeight+8 - inputPad + 8},
                                 {x: inputStartX+tz.w + 2, y: startY + historyHeight+8 - inputPad + fsize.height + 6}],
                             fg);
    } else if (data.focus === 'continue') {
        helpText = 'To select the message input, press Up.';
        this.drawRect(w, (size.w-historyWidth)/2 + inputPad, startY+historyHeight+rowHeight+8, historyWidth - inputPad*2, fsize.height+4, inputColor, true);
    }
    var contText = 'Continue';
    var cz = graphics.textSize(w, contText, normalFontSize);
    graphics.drawText(w, contText, (size.w-cz.w)/2, startY + historyHeight + rowHeight + 8+4 + fsize.height*2/3, normalFontSize, fg);

    var hz = graphics.textSize(w, helpText, normalFontSize);
    graphics.drawText(w, helpText, (size.w-hz.w)/2, startY + historyHeight + rowHeight*2 + 8+4 + fsize.height*2/3, normalFontSize, gray);

    var introText = 'Use this message screen to chat with your partner.';
    var iz = graphics.textSize(w, introText, normalFontSize);
    graphics.drawText(w, introText, (size.w-iz.w)/2, startY - 8+4 - fsize.height*2/3, normalFontSize, fg);
};

GraphicsHelper.prototype.cleanUpActiveQuestion = function() {
    graphics.cleanUpActiveQuestion();
};

GraphicsHelper.prototype.askQuestion = function(data) {
    var question = data.question;
    if (question.type === 'radio') {
        var match;
        for (let i=0; i<question.options.length; i++)
            if( question.options[i][0] === data.answer ) {
                match = i;
                break;
            }
        graphics.askRadioQuestion(question.name, question.query, question.options, !question.disablePrev, match);
    } else if (question.type === 'check') {
        var tmp;
        if (data.answer) tmp = data.answer.split(/,/);
        graphics.askCheckboxQuestion(question.name, question.query, question.options, !question.disablePrev, tmp);
    } else if (question.type === 'oneline') {
        graphics.askOneLineQuestion(question.name, question.query, !question.disablePrev, data.answer);
    } else if (question.type === 'multiline') {
        graphics.askMultiLineQuestion(question.name, question.query, !question.disablePrev, data.answer);
    }
};

// Networking helper functions

function NetworkHelper() {
    this.broadcastCount = 0;
    this.heartBeat = 30;
    this.latency = [0,0];
}

NetworkHelper.prototype = {};

NetworkHelper.prototype.broadcastScreenState = function (state) {
    state.latency = this.latency;

    var s = JSON.stringify(state);
    graphics.serverSend(s);

    this.broadcastCount += 1;
    if (this.broadcastCount % this.heartBeat === 0) {
        var now = new Date().getTime();
        graphics.serverSend(JSON.stringify({screen:"ping", ts: now}));
    }
};

NetworkHelper.prototype.sendScreenState = function (player, state) {
    state.latency = this.latency;

    var s = JSON.stringify(state);
    graphics.serverSend(s, player);
};

NetworkHelper.prototype.waitForClients = function (broadcast) {
    if (broadcast) graphics.broadcastAvailability();
    graphics.checkForNewClients();
};

NetworkHelper.prototype.processClient = function (clientIndex, pnum, keyMapping, inputEvent) {
    var messages = graphics.serverRead(clientIndex);
    if (messages) {
        for (let j=0; j<messages.length; j++) {
            var e = JSON.parse(messages[j]);
            e.player = pnum;

            if (e.type === 'pong') {
                var now = new Date().getTime();
                this.latency[pnum] = now - e.ts;
                e.now = now;
                e.latency = this.latency[pnum];
            }

            inputEvent(e);

            if (e.type === 'keyDown') {
                var k = keyMapping[e.name];
                if (!e.repeat && k) {
                    inputEvent({type: 'gameKeyDown', key: k, player: pnum});
                }
            } else if (e.type === 'keyUp') {
                var k = keyMapping[e.name];
                if (k) {
                    var map = keyMapping[e.name];
                    inputEvent({type: 'gameKeyUp', key: k, player: pnum});
                }
            }
        }
    }
};



exports.GraphicsHelper = GraphicsHelper;
exports.NetworkHelper = NetworkHelper;

// FIXME: Easier way to do this?
exports.initGraphics = graphics.initGraphics;
exports.shutdownGraphics = graphics.shutdownGraphics;
exports.createWindows = graphics.createWindows;
exports.setFullscreen = graphics.setFullscreen;
exports.numWindows = graphics.numWindows;
exports.startDrawing = graphics.startDrawing;
exports.setupGameArea = graphics.setupGameArea;
exports.finishDrawing = graphics.finishDrawing;
exports.flip = graphics.flip;
exports.clipDrawingArea = graphics.clipDrawingArea;
exports.unclipDrawingArea = graphics.unclipDrawingArea;
exports.drawWireFrame = graphics.drawWireFrame;
exports.drawExplosion = graphics.drawExplosion;
exports.drawHexagon = graphics.drawHexagon;
exports.fillHexagon = graphics.fillHexagon;
exports.drawGameArea = graphics.drawGameArea;
exports.drawScore = graphics.drawScore;
exports.drawText = graphics.drawText;
exports.drawImage = graphics.drawImage;
exports.imageSize = graphics.imageSize;
exports.textSize = graphics.textSize;
exports.screenSize = graphics.screenSize;
exports.processEvents = graphics.processEvents;

exports.mainMenu = graphics.mainMenu;

exports.debug = graphics.debug;

exports.startServer = graphics.startServer;
exports.stopServer = graphics.stopServer;
exports.checkForNewClients = graphics.checkForNewClients;
exports.getNumClients = graphics.getNumClients;
exports.broadcastAvailability = graphics.broadcastAvailability;
exports.serverSend = graphics.serverSend;
exports.serverRead = graphics.serverRead;
exports.serverIsConnected = graphics.serverIsConnected;

exports.findServer = graphics.findServer;

exports.startClient = graphics.startClient;
exports.stopClient = graphics.stopClient;
exports.clientSend = graphics.clientSend;
exports.clientRead = graphics.clientRead;
exports.clientIsConnected = graphics.clientIsConnected;
exports.clientSendAudio = graphics.clientSendAudio;
exports.startTextInput = graphics.startTextInput;
exports.stopTextInput = graphics.stopTextInput;
exports.unpauseAudioRecording = graphics.unpauseAudioRecording;
exports.pauseAudioRecording = graphics.pauseAudioRecording;
exports.serverStartLoggingAudio = graphics.serverStartLoggingAudio;
exports.serverStopLoggingAudio = graphics.serverStopLoggingAudio;
