#include "graphics.hh"

WAV::WAV() {
  mBytesWritten = 0;
}

WAV::~WAV() {
  close();
}

void WAV::open(const std::string &filename, int sampleRate, int channels) {
  std::ios::openmode mode = std::ios::out | std::ios::binary | std::ios::trunc;
  mFile.open( filename.c_str(), mode );

  mHeader.sampleRate = sampleRate;
  mHeader.avgBytesPerSecond = sampleRate * 4;
  mHeader.channels = channels;
  mBytesWritten = 0;

  mFile.write( (char *)&mHeader, sizeof( mHeader ));
  mFile.write( (char *)&mFact, sizeof( mFact ));
  mFile.write( (char *)&mData, sizeof( mData ));
}

bool WAV::isOpen() {
  return mFile.is_open();
}

void WAV::close() {
  if( !mFile.is_open()) return;

  mHeader.size = sizeof(mHeader) - 8 + sizeof(mFact) + sizeof(mData) + mBytesWritten;
  mFact.sampleLength = mBytesWritten / 4;
  mData.size = mBytesWritten;

  // Fill in details in the headers
  mFile.seekp(0, std::ios::beg);
  mFile.write( (char *)&mHeader, sizeof( mHeader ));
  mFile.write( (char *)&mFact, sizeof( mFact ));
  mFile.write( (char *)&mData, sizeof( mData ));

  mFile.close();
}

void WAV::write(const char *data, int bytes) {
  if( !mFile.is_open()) return;

  mFile.write(data, bytes);
  mBytesWritten += bytes;
}
