function deg2rad(a) {
    return a * Math.PI / 180;
}

function Wireframe(points, runs, color) {
    Object.call(this);
    this.points = points;
    this.runs = runs;
    this.color = color;
    return this;
}

Wireframe.prototype = {};

Wireframe.prototype.draw = function (render, x, y, angle, color) {
    var points = this.rotateTranslatePoints(x, y, angle);
    render.color = color;
    for (let r=0; r<this.runs.length; r++) {
        var t = new Array(this.runs[r].length);
        for (let i=0; i<this.runs[r].length; i++)
            t[i] = points[this.runs[r][i]];
        render.drawLine(t);
    }
};

Wireframe.prototype.fill = function (render, x, y, angle, color) {
    this.draw(render, x, y, angle, color);
};

Wireframe.prototype.rotateTranslatePoints = function (x, y, angle) {
    var a = deg2rad(angle);
    var c = Math.cos(a);
    var s = Math.sin(a);
    var ret = new Array(this.points.length);
    for (let i=0; i<this.points.length; i++) {
        ret[i] = [this.points[i][0] * c - this.points[i][1] * s + x,
                  this.points[i][0] * s + this.points[i][1] * c + y];
    }
    return ret;
};

Wireframe.prototype.collideCircle = function (p, angle, c, cr) {
    var points = this.rotateTranslatePoints(p, angle);
    for (let i=0; i<points.length; i++) {
        if (distance(points[i], c) <= cr) return true;
    }
    return false;
};

var missileWireframe = new Wireframe([[0, 0], [-25, 0], [-5, 5], [-5, -5]],
                                     [[0,1], [0, 2], [0, 3]],
                                    '#FFFFFF');
var shellWireframe = new Wireframe([[-8, 0], [0, -6], [16, 0], [0, 6]],
                                   [[0,1,2,3,0]],
                                   '#FF0000');
var shipWireframe = new Wireframe([[-18,0], [18,0], [0,0], [-18,18], [-18,-18]],
                                  [[0,1], [3, 2, 4]],
                                  '#FFFF00');
var fortressWireframe = new Wireframe([[0, 0], [36, 0], [18, -18], [0,-18], [18,18], [0,18]],
                                      [[0, 1], [3,2,4,5]],
                                     '#FFFF00');

function drawWireframeExplosion(render, x, y) {
    var ofs = 0;
    var radius;
    for (radius=15; radius<70; radius+=8) {
        var angle;
        ofs += 3;
        if (radius < 60) { render.strokeStyle = '#FFFF00'; }
        else { render.strokeStyle = '#FF0000'; }
        for (angle=0; angle<360; angle += 30) {
            render.beginPath();
            render.arc(x, y, radius, deg2rad(angle+ofs),deg2rad(angle+ofs+10),false);
            render.stroke();
        }
    }
    render.strokeStyle = '#FFFF00';
    render.beginPath();
    render.arc(x, y, 7, 0, Math.PI*2, false);
    render.stroke();
}

exports.missileWireframe = missileWireframe;
exports.shellWireframe = shellWireframe;
exports.shipWireframe = shipWireframe;
exports.fortressWireframe = fortressWireframe;

exports.drawWireframeExplosion = drawWireframeExplosion;
